
DATA_TYPES = {
    'integer': 'Integer',
    'float': 'Float',
    'datetime': 'Date',
    'boolean': 'Boolean',
    'string': 'Text',
    'file': 'File'
}

SCHEMA_OBJS = {
    'entity': 'Entities',
    'relation': 'Relations'
}



from flask import request

from grano.core import db
from grano.model import Permission, Project
from grano.lib.exc import Forbidden


def permissions():
    """ Cache the full matrix of which projects this user has access to. """
    if not hasattr(request, 'permissions'):
        matrix = {
            'reader': set(),
            'editor': set(),
            'admin': set()
        }
        q = db.session.query(Project.id)
        q = q.filter(Project.private == False)
        matrix['reader'].update([id for id, in q.all()])

        if logged_in():
            q = Permission.all()
            q = q.filter_by(account=request.account)
            for perm in q.all():
                if perm.reader:
                    matrix['reader'].add(perm.project_id)
                if perm.editor:
                    matrix['editor'].add(perm.project_id)
                if perm.admin:
                    matrix['admin'].add(perm.project_id)
        request.permissions = matrix
    return request.permissions


def logged_in():
    return request.account is not None


def project_create():
    return logged_in()


def project_read(project):
    return project.id in permissions().get('reader')


def project_edit(project):
    return project.id in permissions().get('editor')


def project_manage(project):
    return project.id in permissions().get('admin')


def project_delete(project):
    return project_manage(project)

# Entity


def entity_create():
    return logged_in()


def entity_read(entity):
    return entity.project_id in permissions().get('reader')


def entity_edit(entity):
    return entity.project_id in permissions().get('editor')


def entity_manage(entity):
    return entity.project_id in permissions().get('admin')


def entity_delete(entity):
    return entity_manage(entity)


def relation_read(relation):
    return relation.project_id in permissions().get('reader')


def relation_edit(relation):
    return relation.project_id in permissions().get('editor')


def relation_manage(relation):
    return relation.project_id in permissions().get('admin')


def require(pred):
    if not pred:
        raise Forbidden("Sorry, you're not permitted to do this!")



from uuid import uuid4
from datetime import datetime

from sqlalchemy.sql.expression import select, func
from sqlalchemy import or_

from grano.core import db
from grano.model import Account, Schema, Entity
from grano.model import BidiRelation
from grano.model.property import Property, VALUE_COLUMNS
from grano.query.parser import EXTRA_FIELDS, EntityParserNode


class Query(object):

    marker = None

    def __init__(self, parent, name, node):
        self.parent = parent
        self.node = node
        self.name = name
        self.results = {}
        if name is None:
            self.id = 'root'
        else:
            prefix = '_any' if name == '*' else name
            self.id = '%s_%s' % (prefix, uuid4().hex[:10])

    def join(self, from_obj, partial=False):
        return from_obj

    def query(self, parent_ids=None):
        pass

    def get_child_node_value(self, name, default=None):
        for node in self.node.children:
            if node.name == name:
                return node.value
        return default

    @property
    def optional(self):
        if self.parent is not None and self.parent.optional:
            return True
        if self.get_child_node_value('optional', False):
            return True
        return False


class FieldQuery(Query):
    """ Query a simple field, as opposed to a more complex nested
    object. """
    
    def __init__(self, parent, name, node):
        super(FieldQuery, self).__init__(parent, name, node)

    @property
    def filtered(self):
        return self.node.value is not None

    @property
    def column(self):
        return getattr(self.parent.alias.c, self.name)

    def filter(self, q, partial=False):
        if self.filtered:
            cond = self.column == self.node.value
            if self.optional:
                cond = or_(cond, self.column==None)
            q = q.where(cond)
        return q

    def project(self, q):
        if self.name == 'id' or not self.filtered:
            q = q.column(self.column.label(self.id))
        return q

    def collect(self, row):
        if self.filtered:
            val = self.node.value
        else:
            val = row.get(self.id, None)
        self.results[row.get(self.parent.pk_id)] = val

    def assemble(self, parent_id):
        return self.results.get(parent_id)


class ObjectQuery(Query):

    model = {}
    domain_object = None
    
    def __init__(self, parent, name, node):
        super(ObjectQuery, self).__init__(parent, name, node)
        self.children = {}
        self.alias = self.domain_object.__table__.alias(self.id)

        # instantiate the model:
        for name, cls in self.model.items():
            for node in self.node.children:
                if node.name == name:
                    self.children[name] = cls(self, name, node)

    @property
    def pk_id(self):
        for name, child in self.children.items():
            if name == 'id':
                return child.id

    @property
    def filtered(self):
        for child in self.node.children:
            if child.name in EXTRA_FIELDS:
                continue
            if self.children[child.name].filtered:
                return True

    def join(self, from_obj, partial=False):
        """ Apply the joins specified on this level of the query. """
        if partial and not self.filtered:
            return from_obj

        if self.parent:
            from_obj = self.join_parent(from_obj)

        for child in self.node.children:
            if child.name in EXTRA_FIELDS:
                continue
            if child.name in self.children:
                field = self.children[child.name]
                from_obj = field.join(from_obj, partial=True)
        return from_obj

    def filter(self, q, partial=False):
        """ Apply the joins specified on this level of the query. """
        if not self.filtered:
            return q

        for child in self.node.children:
            if child.name in EXTRA_FIELDS:
                continue
            if child.name in self.children:
                field = self.children[child.name]
                q = field.filter(q, partial=partial)
        return q

    def join_parent(self, from_obj):
        return from_obj

    def project(self, q):
        """ Define the columns to be retrieved when this is the active
        level of the query. """
        if self.parent is not None:
            q = q.column(self.parent.alias.c.id.label(self.parent.pk_id))

        for name, child in self.children.items():
            for node in self.node.children:
                if not node.name == name:
                    continue
                if isinstance(child, FieldQuery):
                    q = child.project(q)
        return q

    def query(self, parent_ids=None):
        """ Construct a SQL query for this level of the request. """
        if self.parent is None:
            q = select(from_obj=self.join(self.alias))
            q = q.offset(self.get_child_node_value('offset', 0))
            if not self.node.as_list:
                q = q.limit(1)
            else:
                q = q.limit(self.get_child_node_value('limit', 10))
        else:
            q = select(from_obj=self.join(self.parent.alias))
        q = self.filter(q)
        
        if parent_ids is not None:
            q = q.where(self.parent.alias.c.id.in_(parent_ids))
        
        q = self.project(q)
        q = q.distinct()
        #print self, type(self)
        #print q

        ids = []
        rp = db.session.execute(q)
        while True:
            row = rp.fetchone()
            if row is None:
                break
            row = dict(row.items())
            ids.append(row.get(self.pk_id))
            self.collect(row)

        for name, child in self.children.items():
            child.query(parent_ids=ids)

    def count(self):
        """ Get a count of the number of distinct objects. """
        q = select(from_obj=self.join(self.alias))
        q = self.filter(q, partial=True)
        q = q.column(func.count(func.distinct(self.alias.c.id)).label('num'))
        rp = db.session.execute(q)
        return rp.fetchone().num

    def collect(self, row):
        parent_id = row.get(self.parent.pk_id) if self.parent else None
        if parent_id not in self.results:
            self.results[parent_id] = {}
        id = row.get(self.pk_id)
        if id not in self.results[parent_id]:
            data = {}
            if self.marker is not None:
                data['obj'] = self.marker
            self.results[parent_id][id] = data

        for name, child in self.children.items():
            if isinstance(child, FieldQuery):
                child.collect(row)

    def assemble(self, parent_id):
        data = self.results.get(parent_id, {})
        items = []
        for id, item in data.items():
            for name, child in self.children.items():
                item[name] = child.assemble(id)
            items.append(item)

            if not self.node.as_list:
                break

        if not self.node.as_list and len(items):
            return items.pop()
        return items

    def run(self):
        self.query()
        return self.assemble(None)


class IdFieldQuery(FieldQuery):

    def assemble(self, parent_id):
        id = super(IdFieldQuery, self).assemble(parent_id)
        return id.split('-', 1)[0]


class AuthorQuery(ObjectQuery):
    
    domain_object = Account
    model = {
        'id': FieldQuery,
        'login': FieldQuery,
        'full_name': FieldQuery,
        'created_at': FieldQuery,
        'updated_at': FieldQuery
    }

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.id == self.parent.alias.c.author_id,
                             isouter=self.optional)


class SchemaQuery(ObjectQuery):
    
    domain_object = Schema
    model = {
        'id': FieldQuery,
        'name': FieldQuery,
        'hidden': FieldQuery,
        'label': FieldQuery,
        'created_at': FieldQuery,
        'updated_at': FieldQuery
    }

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.id == self.parent.alias.c.schema_id,
                             isouter=self.optional)


class PropertyQuery(ObjectQuery):
    """ Property queries are the second level in querying a set of
    properties, they are called by the PropertiesQuery. This is somewhat
    complex because we need to handle types (i.e. query the appropriate
    column for the submitted input type, or retrieve all to find the
    one that holds a value. """

    domain_object = Property
    model = {
        'id': FieldQuery,
        'name': FieldQuery,
        'value_string': FieldQuery,
        'value_datetime': FieldQuery,
        'value_integer': FieldQuery,
        'value_float': FieldQuery,
        'value_boolean': FieldQuery,
        'source_url': FieldQuery,
        'active': FieldQuery
    }

    def __init__(self, parent, name, node):
        if 'value' in node.value:
            obj = node.value.pop('value')
            if obj is None:
                for col in VALUE_COLUMNS:
                    node.value[col] = None
            else:
                node.value[Property.type_column(obj)] = obj

        if name == '*':
            node.value['name'] = None
        else:
            node.value['name'] = name
        node.value['active'] = True
        node.value['optional'] = True
        node.as_list = True
        super(PropertyQuery, self).__init__(parent, name, node)

    def filter(self, q, partial=False):
        for child in self.node.children:
            if child.name in EXTRA_FIELDS:
                continue
            if child.name in self.children:
                field = self.children[child.name]
                q = field.filter(q, partial=partial)
        return q

    @property
    def filtered(self):
        for name, child in self.children.items():
            if child.filtered:
                if name in VALUE_COLUMNS:
                    return True
            #if name == 'name' and child.filtered:
            #    return True
        return False

    def assemble(self, parent_id):
        props = {}
        for prop in super(PropertyQuery, self).assemble(parent_id):
            prop['value'] = None
            for col in VALUE_COLUMNS:
                col_val = prop.pop(col, None)
                if col_val is not None:
                    prop['value'] = col_val
            props[prop.pop('name')] = prop
        return props


class EntityPropertyQuery(PropertyQuery):

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.entity_id == self.parent.alias.c.id,
                             isouter=self.optional)


class RelationPropertyQuery(PropertyQuery):

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.relation_id == self.parent.alias.c.relation_id,
                             isouter=self.optional)


class PropertiesQuery(object):
    """ A stub query object to retrieve all the requested properties
    and return them in an associative array. """

    def __init__(self, parent, name, node):
        self.parent = parent
        self.children = []
        for child in node.children:
            prop = self.child_cls(parent, child.name, child)
            self.children.append(prop)

    @property
    def filtered(self):
        for child in self.children:
            if child.filtered:
                return True
        return False

    def filter(self, q, partial=False):
        for child in self.children:
            if child.filtered:
                q = child.filter(q, partial=partial)
        return q

    def join(self, from_obj, partial=False):
        for child in self.children:
            from_obj = child.join(from_obj, partial=partial)
        return from_obj

    def query(self, parent_ids=None):
        for child in self.children:
            child.query(parent_ids=parent_ids)

    def collect(self, row):
        for child in self.children:
            child.collect(row)

    def assemble(self, parent_id):
        data = {}
        for child in self.children:
            data.update(child.assemble(parent_id))
        return data

    @property
    def optional(self):
        return True


class EntityPropertiesQuery(PropertiesQuery):
    child_cls = EntityPropertyQuery


class RelationPropertiesQuery(PropertiesQuery):
    child_cls = RelationPropertyQuery


class RelationQuery(ObjectQuery):
    
    domain_object = BidiRelation
    marker = 'relation'
    model = {
        'id': IdFieldQuery,
        'reverse': FieldQuery,
        'author': AuthorQuery,
        'schema': SchemaQuery,
        'properties': RelationPropertiesQuery,
        'created_at': FieldQuery,
        'updated_at': FieldQuery
    }

    @property
    def filtered(self):
        return True

    def bidi_filter(self, q):
        return q.where(self.alias.c.reverse == False) # noqa

    def filter(self, q, partial=False):
        if self.filtered or not partial:
            q = q.where(self.alias.c.project_id == self.node.project.id)
            q = self.bidi_filter(q)
        return super(RelationQuery, self).filter(q, partial=partial)


class InboundRelationQuery(RelationQuery):

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.target_id == self.parent.alias.c.id,
                             isouter=self.optional)


class OutboundRelationQuery(RelationQuery):

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.source_id == self.parent.alias.c.id,
                             isouter=self.optional)


class BidiRelationQuery(RelationQuery):

    def bidi_filter(self, q):
        return q

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.target_id == self.parent.alias.c.id,
                             isouter=self.optional)


class EntityQuery(ObjectQuery):
    
    domain_object = Entity
    marker = 'entity'
    model = {
        'id': FieldQuery,
        'created_at': FieldQuery,
        'updated_at': FieldQuery,
        'degree_in': FieldQuery,
        'degree_out': FieldQuery,
        'degree': FieldQuery,
        'schema': SchemaQuery,
        'author': AuthorQuery,
        'inbound': InboundRelationQuery,
        'outbound': OutboundRelationQuery,
        'relations': BidiRelationQuery,
        'properties': EntityPropertiesQuery,
    }

    @property
    def filtered(self):
        return True

    def filter(self, q, partial=False):
        if self.filtered or not partial or not self.parent:
            q = q.where(self.alias.c.project_id == self.node.project.id)
            q = q.where(self.alias.c.same_as == None) # noqa
        return super(EntityQuery, self).filter(q, partial=partial)


class SourceEntityQuery(EntityQuery):

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.id == self.parent.alias.c.source_id,
                             isouter=self.optional)


class TargetEntityQuery(EntityQuery):

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.id == self.parent.alias.c.target_id,
                             isouter=self.optional)


class BidiEntityQuery(EntityQuery):

    def join_parent(self, from_obj):
        return from_obj.join(self.alias,
                             onclause=self.alias.c.id == self.parent.alias.c.source_id,
                             isouter=self.optional)


InboundRelationQuery.model['source'] = SourceEntityQuery
OutboundRelationQuery.model['target'] = TargetEntityQuery
BidiRelationQuery.model['other'] = BidiEntityQuery


def run_query(project, query):
    node = EntityParserNode(project, None, query)
    node.value['limit'] = min(1000, node.value.get('limit', 25))
    node.value['offset'] = max(0, node.value.get('offset', 0))
    eq = EntityQuery(None, None, node)
    return eq



from flask import request
from sqlalchemy.orm import aliased
from sqlalchemy import or_

from grano.authz import permissions
from grano.model import Project, Relation
from grano.model import Entity, Property, Schema
from grano.lib.args import single_arg


PROPERTY = 'property-'
ALIASES = 'aliases-'


def property_filters(cls, q, alias):
    """ Parse the query arguments and apply any specified property
    filters to the given query ``q``. The property-holding object
    (a relation or entity) is given as ``cls``. """
    for key in request.args.keys():
        if not key.startswith(PROPERTY):
            continue
        prop = key[len(PROPERTY):]

        only_active = True
        if prop.startswith(ALIASES):
            prop = prop[len(ALIASES):]
            only_active = False

        value = single_arg(key)
        q = cls._filter_property(q, prop, value,
                                 only_active=only_active,
                                 alias=alias)
    return q


def for_relations(q, Rel):
    #Source = aliased(Entity)
    #Target = aliased(Entity)
    #q = q.join(Source, Rel.source)
    #q = q.join(Target, Rel.target)

    q = q.filter(Rel.project_id.in_(permissions().get('reader')))

    project = single_arg('project')
    if project:
        Proj = aliased(Project)
        q = q.join(Proj, Rel.project)
        q = q.filter(Proj.slug == project)

    q = property_filters(Relation, q, Rel)

    if 'source' in request.args:
        q = q.filter(Rel.source_id == single_arg('source'))

    if 'target' in request.args:
        q = q.filter(Rel.target_id == single_arg('target'))

    if 'entity' in request.args:
        id = single_arg('entity')
        q = q.filter(or_(Rel.target_id == id, Rel.source_id == id))

    schemata = request.args.getlist('schema')
    if len(schemata):
        alias = aliased(Schema)
        q = q.join(alias, Rel.schema)
        q = q.filter(alias.name.in_(schemata))

    return q


def for_entities(q, Ent):
    """ Get all entities the current user has access to. Accepts project and
    additional filter parameters. """
    # NOTE: I'm passing in the query and entity alias so that this
    # function can be re-used from the facetting code to constrain
    # the results of the facet sub-query.
    q = q.filter(Ent.same_as == None) # noqa
    q = q.filter(Ent.project_id.in_(permissions().get('reader')))

    if 'project' in request.args:
        Proj = aliased(Project)
        q = q.join(Proj, Ent.project)
        q = q.filter(Proj.slug == single_arg('project'))

    q = property_filters(Entity, q, Ent)

    if 'q' in request.args and single_arg('q'):
        EntProp = aliased(Property)
        q_text = '%%%s%%' % single_arg('q')
        q = q.join(EntProp)
        q = q.filter(EntProp.name == 'name')
        q = q.filter(EntProp.value_string.ilike(q_text))

    schemata = request.args.getlist('schema')
    if len(schemata):
        alias = aliased(Schema)
        q = q.join(alias, Ent.schema)
        q = q.filter(alias.name.in_(schemata))

    return q



from sqlalchemy import func, cast, types
from sqlalchemy.orm import aliased

from grano.interface import Startup
from grano.model import Entity, Property, Schema
from grano.core import db


class Matches(object):

    def __init__(self, q, project, account):
        self.lq = self.q = q
        self.account = account
        self.project = project

    def limit(self, l):
        self.lq = self.lq.limit(l)
        return self

    def offset(self, o):
        self.lq = self.lq.offset(o)
        return self

    def count(self):
        return self.q.count()

    def __iter__(self):
        rows = self.lq.all()
        ids = [r[0] for r in rows]
        q = db.session.query(Entity)
        q = q.filter(Entity.project == self.project)
        q = q.filter(Entity.id.in_(ids))
        by_id = {}
        for entity in q.all():
            by_id[entity.id] = entity
        for (id, score) in rows:
            yield {
                'score': int(score),
                'entity': by_id.get(id)
            }


def find_matches(project, account, text, schemata=[], properties=[]):
    main = aliased(Property)
    ent = aliased(Entity)
    q = db.session.query(main.entity_id)
    q = q.filter(main.name == 'name')
    q = q.filter(main.entity_id == ent.id)
    q = q.join(ent)
    q = q.filter(ent.project_id == project.id)

    if len(schemata):
        obj = aliased(Schema)
        q = q.join(obj, ent.schema_id == obj.id)
        q = q.filter(obj.name.in_(schemata))

    for name, value in properties:
        p = aliased(Property)
        q = q.join(p, p.entity_id == ent.id)
        q = q.filter(p.active == True) # noqa
        q = q.filter(p.name == name)
        column = getattr(p, p.type_column(value))
        q = q.filter(column == value)

    # prepare text fields (todo: further normalization!)
    text_field = func.left(func.lower(main.value_string), 254)
    match_text = text.lower().strip()[:254]
    match_text_db = cast(match_text, types.Unicode)

    # calculate the difference percentage
    l = func.greatest(1.0, func.least(len(match_text), func.length(text_field)))
    score = func.greatest(0.0, ((l - func.levenshtein(text_field, match_text_db)) / l) * 100.0)
    score = score.label('score')
    q = q.group_by(main.entity_id)
    q = q.add_columns(func.max(score))
    q = q.order_by(func.max(score).desc())
    q = q.filter(score > 50)
    return Matches(q, project, account)


class ConfigurePostgres(Startup):

    def configure(self, manager):
        ext = "CREATE EXTENSION IF NOT EXISTS fuzzystrmatch;"
        db.session.execute(ext)
        db.session.commit()



from sqlalchemy import func

from grano.core import db, url_for
from grano.model.util import MutableDict, JSONEncodedDict
from grano.model.common import IntBase


class Project(db.Model, IntBase):
    __tablename__ = 'grano_project'

    slug = db.Column(db.Unicode)
    label = db.Column(db.Unicode)
    private = db.Column(db.Boolean, default=False)
    settings = db.Column(MutableDict.as_mutable(JSONEncodedDict))

    author_id = db.Column(db.Integer, db.ForeignKey('grano_account.id'))

    relations = db.relationship('Relation', backref='project', lazy='dynamic',
                                cascade='all, delete, delete-orphan')
    entities = db.relationship('Entity', backref='project', lazy='dynamic',
                               cascade='all, delete, delete-orphan')
    pipelines = db.relationship('Pipeline', backref='project', lazy='dynamic',
                                cascade='all, delete, delete-orphan')
    schemata = db.relationship('Schema', backref='project', lazy='dynamic',
                               cascade='all, delete, delete-orphan')
    permissions = db.relationship('Permission', backref='project',
                                  lazy='dynamic',
                                  cascade='all, delete, delete-orphan')
    files = db.relationship('File', backref='project', lazy='dynamic',
                            cascade='all, delete, delete-orphan')

    @classmethod
    def by_slug(cls, slug):
        q = db.session.query(cls)
        q = q.filter(func.lower(cls.slug) == slug.lower())
        return q.first()

    def to_dict_short(self):
        return {
            'slug': self.slug,
            'label': self.label,
            'private': self.private,
            'api_url': url_for('projects_api.view', slug=self.slug)
        }

    def to_dict_index(self):
        data = self.to_dict_short()
        data['entities_count'] = self.entities.count()
        data['relations_count'] = self.relations.count()
        return data

    def to_dict(self):
        data = self.to_dict_index()
        data['settings'] = self.settings
        data['author'] = self.author.to_dict_index()
        data['schemata_index_url'] = url_for('schemata_api.index',
                                             slug=self.slug)
        data['entities_index_url'] = url_for('entities_api.index',
                                             project=self.slug)
        data['relations_index_url'] = url_for('relations_api.index',
                                              project=self.slug)
        return data



from StringIO import StringIO
from unicodecsv import DictReader
from unicodecsv import Error as CSVError
import logging
import colander

from grano.core import db
from grano.model import File
from grano.logic.references import ProjectRef, AccountRef
from grano.lib.exc import BadRequest


log = logging.getLogger(__name__)


class FileValidator(colander.MappingSchema):
    author = colander.SchemaNode(AccountRef())
    project = colander.SchemaNode(ProjectRef())
    file_name = colander.SchemaNode(colander.String(),
                                    validator=colander.Length(min=3))
    mime_type = colander.SchemaNode(colander.String(),
                                    validator=colander.Length(min=3))


def validate(data, file):
    validator = FileValidator()
    sane = validator.deserialize(data)
    return sane


def save(data, file_data, file=None):
    """ Save or update a file. """
    if file_data is None:
        raise BadRequest("No file given!")

    data.update({
        'file_name': file_data.filename,
        'mime_type': file_data.mimetype
        })
    sane = validate(data, file)

    if file is None:
        file = File()
        file.project = sane.get('project')
        file.author = sane.get('author')
        db.session.add(file)

    file.file_name = sane.get('file_name')
    file.mime_type = sane.get('mime_type')
    file.data = file_data.read()

    db.session.flush()
    return file


def delete(file):
    """ Delete the file. """
    db.session.delete(file)


def as_table(file, limit=None):
    try:
        sio = StringIO(file.data)
        reader = DictReader(sio)
        data = {'headers': None, 'rows': [], 'total': 0}
        for i, row in enumerate(reader):
            if data['headers'] is None:
                data['headers'] = row.keys()
            if limit is None or i < limit:
                rd = [row.get(k) for k in data['headers']]
                data['rows'].append(rd)
            data['total'] = i
        return data
    except CSVError, e:
        return {'status': 'error', 'error': unicode(e)}



import logging

from grano.core import celery as app
from grano.plugins import notify_plugins
from grano.logic.imports import run_importer # noqa


log = logging.getLogger(__name__)


@app.task
def ping(message):
    log.warning("Ping was registered, message: %s", message)


@app.task
def periodic():
    def _handle(obj):
        obj.run()
    notify_plugins('grano.periodic', _handle)



import requests
from flask import session, Blueprint, redirect
from flask import request


from grano import authz
from grano.lib.exc import BadRequest
from grano.lib.serialisation import jsonify
from grano.views.cache import validate_cache
from grano.core import db, url_for
from grano.providers import PROVIDERS, Stub
from grano.model import Account
from grano.logic import accounts


blueprint = Blueprint('sessions_api', __name__)


@blueprint.route('/api/1/sessions', methods=['GET'])
def status():
    permissions = {}
    if authz.logged_in():
        for permission in request.account.permissions:
            permissions[permission.project.slug] = {
                'reader': permission.reader,
                'editor': permission.editor,
                'admin': permission.admin
            }

    keys = {
        'p': repr(permissions),
        'i': request.account.id if authz.logged_in() else None
    }
    validate_cache(keys=keys)

    oauth_providers = {}
    for name, provider in PROVIDERS.items():
        if not isinstance(provider, Stub):
            oauth_providers[name] = url_for('.login', provider=name)

    return jsonify({
        'logged_in': authz.logged_in(),
        'api_key': request.account.api_key if authz.logged_in() else None,
        'account': request.account if request.account else None,
        'permissions': permissions,
        'oauth': oauth_providers
    })


@blueprint.route('/api/1/sessions/logout', methods=['GET'])
def logout():
    session.clear()
    return redirect(request.args.get('next_url', '/'))


@blueprint.route('/api/1/sessions/login/<provider>', methods=['GET'])
def login(provider):
    if provider not in PROVIDERS:
        raise BadRequest('Unknown provider: %s' % provider)
    callback = url_for('sessions_api.%s_authorized' % provider)
    session.clear()
    if not request.args.get('next_url'):
        raise BadRequest("No 'next_url' is specified.")
    session['next_url'] = request.args.get('next_url')
    return PROVIDERS[provider].authorize(callback=callback)


handler = PROVIDERS.get('github')


@blueprint.route('/api/1/sessions/callback/github', methods=['GET'])
@handler.authorized_handler
def github_authorized(resp):
    next_url = session.get('next_url', '/')
    if resp is None or 'access_token' not in resp:
        return redirect(next_url)
    access_token = resp['access_token']
    session['access_token'] = access_token, ''
    url = 'https://api.github.com/user?access_token=%s'
    res = requests.get(url % access_token, verify=False)
    data = res.json()
    account = Account.by_github_id(data.get('id'))
    data_ = {
        'full_name': data.get('name'),
        'login': data.get('login'),
        'email': data.get('email'),
        'github_id': data.get('id')
    }
    account = accounts.save(data_, account=account)
    db.session.commit()
    session['id'] = account.id
    return redirect(next_url)


handler = PROVIDERS.get('twitter')


@blueprint.route('/api/1/sessions/callback/twitter', methods=['GET'])
@handler.authorized_handler
def twitter_authorized(resp):
    next_url = session.get('next_url', '/')
    if resp is None or 'oauth_token' not in resp:
        return redirect(next_url)
    session['twitter_token'] = (resp['oauth_token'],
                                resp['oauth_token_secret'])
    provider = PROVIDERS.get('twitter')
    res = provider.get('users/show.json?user_id=%s' % resp.get('user_id'))
    account = Account.by_twitter_id(res.data.get('id'))
    data_ = {
        'full_name': res.data.get('name'),
        'login': res.data.get('screen_name'),
        'twitter_id': res.data.get('id')
    }
    account = accounts.save(data_, account=account)
    db.session.commit()
    session['id'] = account.id
    return redirect(next_url)


handler = PROVIDERS.get('facebook')


@blueprint.route('/api/1/sessions/callback/facebook', methods=['GET'])
@handler.authorized_handler
def facebook_authorized(resp):
    next_url = session.get('next_url', '/')
    if resp is None or 'access_token' not in resp:
        return redirect(next_url)
    session['facebook_token'] = (resp.get('access_token'), '')
    data = PROVIDERS.get('facebook').get('/me').data
    account = Account.by_facebook_id(data.get('id'))
    data_ = {
        'full_name': data.get('name'),
        'login': data.get('username'),
        'email': data.get('email'),
        'facebook_id': data.get('id')
    }
    account = accounts.save(data_, account=account)
    db.session.commit()
    session['id'] = account.id
    return redirect(next_url)



from functools import wraps
import certifi

from flask import session

from grano.core import oauth, app
from grano.lib.serialisation import jsonify


class Stub():
    """ A stub authorization handler to sit in for auth methods that
    are not currently enabled. """

    def __init__(self, name):
        self.name = name

    def authorize(self, **kwargs):
        return jsonify({
            'status': 501,
            'name': 'Provider not configured: %s' % self.name,
            'message': 'There are no credentials given for %s' % self.name,
            }, status=501)

    def authorized_handler(self, f):
        @wraps(f)
        def inner(*a, **kw):
            return self.authorize()
        return inner


PROVIDERS = {
    'github': Stub('github'),
    'twitter': Stub('twitter'),
    'facebook': Stub('facebook')
}


if app.config.get('GITHUB_CLIENT_ID') is not None:
    PROVIDERS['github'] = oauth.remote_app('github',
        base_url='https://github.com/login/oauth/',
        authorize_url='https://github.com/login/oauth/authorize',
        request_token_url=None,
        access_token_url='https://github.com/login/oauth/access_token',
        consumer_key=app.config.get('GITHUB_CLIENT_ID'),
        consumer_secret=app.config.get('GITHUB_CLIENT_SECRET'))
    PROVIDERS['github']._client.ca_certs = certifi.where()

if app.config.get('TWITTER_API_KEY') is not None:
    twitter = oauth.remote_app('twitter',
        base_url='https://api.twitter.com/1.1/',
        request_token_url='https://api.twitter.com/oauth/request_token',
        access_token_url='https://api.twitter.com/oauth/access_token',
        authorize_url='https://api.twitter.com/oauth/authenticate',
        consumer_key=app.config.get('TWITTER_API_KEY'),
        consumer_secret=app.config.get('TWITTER_API_SECRET'))

    @twitter.tokengetter
    def get_twitter_token(token=None):
        return session.get('twitter_token')

    PROVIDERS['twitter'] = twitter

if app.config.get('FACEBOOK_APP_ID') is not None:
    facebook = oauth.remote_app('facebook',
        base_url='https://graph.facebook.com/',
        request_token_url=None,
        access_token_url='/oauth/access_token',
        authorize_url='https://www.facebook.com/dialog/oauth',
        consumer_key=app.config.get('FACEBOOK_APP_ID'),
        consumer_secret=app.config.get('FACEBOOK_APP_SECRET'),
        request_token_params={'scope': 'email'})

    @facebook.tokengetter
    def get_facebook_token(token=None):
        return session.get('facebook_token')

    PROVIDERS['facebook'] = facebook



from flask import Blueprint, request
from restpager import Pager

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, request_data
from grano.model import Permission, Project
from grano.views.cache import validate_cache
from grano.logic import permissions
from grano.lib.exc import Gone
from grano.core import db
from grano import authz


blueprint = Blueprint('permissions_api', __name__)


@blueprint.route('/api/1/projects/<slug>/permissions', methods=['GET'])
def index(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    query = Permission.all()
    query = query.filter_by(project=project)
    pager = Pager(query, slug=slug)
    validate_cache(keys=pager.cache_keys())
    return jsonify(pager, index=True)


@blueprint.route('/api/1/projects/<slug>/permissions', methods=['POST', 'PUT'])
def create(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    data = request_data({'project': project})
    permission = permissions.save(data)
    db.session.commit()
    return jsonify(permission, status=201)


@blueprint.route('/api/1/projects/<slug>/permissions/<id>', methods=['GET'])
def view(slug, id):
    project = object_or_404(Project.by_slug(slug))
    permission = object_or_404(Permission.by_project_and_id(project, id))
    authz.require(authz.project_manage(project) or
                  request.account == permission.account)
    return jsonify(permission)


@blueprint.route('/api/1/projects/<slug>/permissions/<id>',
                 methods=['POST', 'PUT'])
def update(slug, id):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    permission = object_or_404(Permission.by_project_and_id(project, id))
    data = request_data({'project': project})
    permission = permissions.save(data, permission=permission)
    db.session.commit()
    return jsonify(permission)


@blueprint.route('/api/1/projects/<slug>/permissions/<id>', methods=['DELETE'])
def delete(slug, id):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    permission = object_or_404(Permission.by_project_and_id(project, id))
    permissions.delete(permission)
    db.session.commit()
    raise Gone()



from flask import request
from sqlalchemy import desc, asc


def parse_sorts():
    for key in request.args.getlist('sort'):
        direction = asc
        if key.startswith('-'):
            direction = desc
            key = key[1:]
        yield key, direction


def for_relations(q, Rel):
    for key, direction in parse_sorts():
        for cand in ['id', 'created_at', 'updated_at']:
            if key == cand:
                q = q.order_by(direction(getattr(Rel, key)))
        if key == 'source':
            q = q.order_by(direction(Rel.source_id, key))
        if key == 'target':
            q = q.order_by(direction(Rel.target_id, key))
    return q


def for_entities(q, Ent):
    for key, direction in parse_sorts():
        for cand in ['id', 'created_at', 'updated_at', 'degree', 'degree_in', 'degree_out']:
            if key == cand:
                q = q.order_by(direction(getattr(Ent, key)))
    return q
    



"""empty message

Revision ID: 4fb92f3ede9e
Revises: 2b269c4557de
Create Date: 2014-03-14 12:12:41.912515

"""

# revision identifiers, used by Alembic.
revision = '4fb92f3ede9e'
down_revision = '2b269c4557de'

from alembic import op
from datetime import datetime
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('permission',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('reader', sa.Boolean(), nullable=True),
    sa.Column('editor', sa.Boolean(), nullable=True),
    sa.Column('admin', sa.Boolean(), nullable=True),
    sa.Column('account_id', sa.Integer(), nullable=True),
    sa.Column('project_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['account_id'], ['account.id'], ),
    sa.ForeignKeyConstraint(['project_id'], ['project.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    #op.drop_table('role')
    ### end Alembic commands ###

    projects = sa.sql.table('project',
        sa.sql.column('id', sa.Integer),
        sa.sql.column('author_id', sa.Integer)
    )

    perms = sa.sql.table('permission',
        sa.sql.column('project_id', sa.Integer),
        sa.sql.column('account_id', sa.Integer),
        sa.sql.column('reader', sa.Boolean),
        sa.sql.column('editor', sa.Boolean),
        sa.sql.column('admin', sa.Boolean),
        sa.sql.column('created_at', sa.DateTime),
        sa.sql.column('updated_at', sa.DateTime)
    )
    
    
    connection = op.get_bind()
    rp = connection.execute(projects.select())

    if rp is not None:
        perm_seed = []
        for row in rp.fetchall():
            row = dict(zip(row.keys(), tuple(row)))
            perm_seed.append({
                'account_id': row.get('author_id'),
                'project_id': row.get('id'),
                'admin': True,
                'reader': True,
                'editor': True,
                'created_at': datetime.utcnow(),
                'updated_at': datetime.utcnow()
                })

        op.bulk_insert(perms, perm_seed)


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('permission')
    ### end Alembic commands ###



from grano.logic.schemata import import_schema, export_schema
from grano.logic.aliases import import_aliases, export_aliases
from grano.logic.loader import Loader
from grano.logic.plugins import rebuild



EXTRA_FIELDS = ['limit', 'offset', 'sort', 'optional', 'obj']


class ParserNode(object):
    """ A query parser object. This will help the query builder traverse
    the request and also fix up elements of the request that are
    ambiguous. """

    child_types = {}
    defaults = None
    key_field = None

    def __init__(self, project, name, source):
        self.project = project
        self.as_list = isinstance(source, (list, tuple))
        self._value = source
        if self.as_list:
            if len(source):
                self._value = source[0]
            else:
                self._value = None
        self.name = name

    @property
    def is_stub(self):
        if self._value is None:
            return True
        if not isinstance(self._value, dict):
            return False
        for k, v in self._value.items():
            if k not in EXTRA_FIELDS:
                return False
        return True

    @property
    def value(self):
        """ This will patch up the object with parsing defaults. """
        if self.is_stub:
            if self.defaults is None:
                return None
            self._value = {'*': None}

        # Transform: "foo" -> {"key": "foo"}
        if self.key_field and isinstance(self._value, basestring):
            self._value = {self.key_field: self._value}

        if isinstance(self._value, dict):

            # Expand wildcard queries
            if '*' in self._value:
                self._value.pop('*')
                for k, v in self.defaults.items():
                    if k not in self._value:
                        self._value[k] = v

            # Make sure ID is always fetched
            if 'id' not in self._value and self.name != 'properties':
                self._value['id'] = None
        return self._value

    @value.setter
    def value(self, value):
        self._value = value

    @property
    def is_leaf(self):
        return not isinstance(self.value, dict)

    @property
    def children(self):
        if self.is_leaf:
            return
        for k, v in self.value.items():
            cls = self.child_types.get(None, ParserNode)
            cls = self.child_types.get(k, cls)
            yield cls(self.project, k, v)

    def to_dict(self):
        if self.is_leaf:
            return self.value
        else:
            data = dict([(c.name, c) for c in self.children])
            return [data] if self.as_list else data


class AccountParserNode(ParserNode):

    defaults = {
        'login': None
    }
    key_field = 'login'


class ProjectParserNode(ParserNode):

    defaults = {
        'slug': None,
        'label': None
    }
    key_field = 'slug'


class SchemaParserNode(ParserNode):

    defaults = {
        'name': None,
        'label': None,
        'hidden': None
    }
    key_field = 'name'


class PropertyParserNode(ParserNode):

    defaults = {
        'value': None,
        'source_url': None
    }
    key_field = 'value'


class PropertiesParserNode(ParserNode):

    defaults = {
        '*': None,
    }
    child_types = {
        None: PropertyParserNode
    }


class EntityParserNode(ParserNode):

    defaults = {
        'id': None,
        'degree': None,
        'schema': {},
        'properties': {}
    }
    child_types = {
        'author': AccountParserNode,
        'project': ProjectParserNode,
        'schema': SchemaParserNode,
        'properties': PropertiesParserNode
    }


class RelationParserNode(ParserNode):

    defaults = {
        'id': None,
        'reverse': None,
        'schema': [],
        'properties': {}
    }
    child_types = {
        'author': AccountParserNode,
        'project': ProjectParserNode,
        'schema': SchemaParserNode,
        'source': EntityParserNode,
        'other': EntityParserNode,
        'target': EntityParserNode,
        'properties': PropertiesParserNode
    }

EntityParserNode.child_types['inbound'] = RelationParserNode
EntityParserNode.child_types['outbound'] = RelationParserNode
EntityParserNode.child_types['relations'] = RelationParserNode



"""remove status again

Revision ID: 59d7b4f94cdf
Revises: 5143617349b9
Create Date: 2014-09-11 20:28:20.938462

"""

# revision identifiers, used by Alembic.
revision = '59d7b4f94cdf'
down_revision = '5143617349b9'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_column(u'grano_entity', 'status')


def downgrade():
    op.add_column(u'grano_entity', sa.Column('status', sa.INTEGER(), nullable=True))



from grano.core import db

from grano.model.schema import Schema
from grano.model.attribute import Attribute
from grano.model.property import Property, DATETIME_PRECISION
from grano.model.entity import Entity
from grano.model.relation import Relation, BidiRelation
from grano.model.account import Account
from grano.model.project import Project
from grano.model.permission import Permission
from grano.model.file import File
from grano.model.pipeline import Pipeline, LogEntry



import logging

from flask_script import Manager
from flask_migrate import MigrateCommand

from grano.views import app
from grano.model import Project
from grano.logic import import_schema, export_schema
from grano.logic import import_aliases, export_aliases
from grano.logic import rebuild as rebuild_
from grano.logic.accounts import console_account
from grano.logic.projects import save as save_project
from grano.plugins import list_plugins, notify_plugins
from grano.background import periodic as periodic_task


log = logging.getLogger('grano')
manager = Manager(app)
manager.add_command('db', MigrateCommand)
notify_plugins('grano.startup', lambda o: o.configure(manager))


@manager.command
def schema_import(project, path):
    """ Load a schema specification from a YAML file. """
    pobj = Project.by_slug(project)
    if pobj is None:
        pobj = save_project({
            'slug': project,
            'label': project,
            'author': console_account()
            })
    with open(path, 'r') as fh:
        import_schema(pobj, fh)


@manager.command
def schema_export(project, path):
    """ Export the current schema to a YAML file. """
    pobj = Project.by_slug(project)
    assert pobj is not None, 'Project not available: %s' % project
    export_schema(pobj, path)


@manager.command
def alias_import(project, path):
    """ Load a set of entity aliases from a CSV file. """
    pobj = Project.by_slug(project)
    assert pobj is not None, 'Project not available: %s' % project
    import_aliases(pobj, console_account(), path)


@manager.command
def alias_export(project, path):
    """ Export all known entity aliases to a CSV file. """
    pobj = Project.by_slug(project)
    assert pobj is not None, 'Project not available: %s' % project
    with open(path, 'w') as fh:
        export_aliases(pobj, fh)


@manager.command
def rebuild():
    """ Trigger change processing on all relations and entities. """
    rebuild_()


@manager.command
def periodic():
    """ Trigger the periodic background service. """
    periodic_task()


@manager.command
def adminkey():
    """ Print the API key for the system account. """
    print console_account().api_key


@manager.command
def plugins():
    """ List all available plugins. """
    for namespace, plugins in list_plugins().items():
        print "%s: %s" % (namespace, ' '.join(plugins))


def run():
    manager.run()

if __name__ == "__main__":
    run()



from datetime import datetime

from grano.core import db
from grano.model.util import make_token


class _CoreBase(object):
    created_at = db.Column(db.DateTime, default=datetime.utcnow)
    updated_at = db.Column(db.DateTime, default=datetime.utcnow,
                           onupdate=datetime.utcnow)

    @classmethod
    def by_id(cls, id):
        q = db.session.query(cls).filter_by(id=id)
        return q.first()

    @classmethod
    def all(cls):
        return db.session.query(cls)


class IntBase(_CoreBase):
    id = db.Column(db.Integer, primary_key=True)

    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, self.id)


class UUIDBase(_CoreBase):
    id = db.Column(db.Unicode, default=make_token, primary_key=True)

    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, self.id)



"""init

Revision ID: 4f21a77e91be
Revises: None
Create Date: 2014-02-13 17:05:34.953371

"""

# revision identifiers, used by Alembic.
revision = '4f21a77e91be'
down_revision = None

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    op.create_table('account',
    sa.Column('created_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('updated_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.INTEGER(), primary_key=True),
    sa.Column('github_id', sa.INTEGER(), autoincrement=False, nullable=True),
    sa.Column('login', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('email', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('api_key', sa.VARCHAR(), autoincrement=False, nullable=True)
    )
    op.create_table('entity',
    sa.Column('created_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('updated_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.VARCHAR(), autoincrement=False, nullable=False),
    sa.Column('same_as', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.ForeignKeyConstraint(['same_as'], [u'entity.id'], name=u'entity_same_as_fkey'),
    sa.PrimaryKeyConstraint('id', name=u'entity_pkey')
    )
    op.create_table('schema',
    sa.Column('created_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('updated_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.INTEGER(), primary_key=True),
    sa.Column('name', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('label', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('label_in', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('label_out', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('obj', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('hidden', sa.BOOLEAN(), autoincrement=False, nullable=True),
    sa.PrimaryKeyConstraint('id', name=u'schema_pkey')
    )
    op.create_table('attribute',
    sa.Column('created_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('updated_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.INTEGER(), primary_key=True),
    sa.Column('name', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('label', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('description', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('hidden', sa.BOOLEAN(), autoincrement=False, nullable=True),
    sa.Column('schema_id', sa.INTEGER(), autoincrement=False, nullable=True),
    sa.ForeignKeyConstraint(['schema_id'], [u'schema.id'], name=u'attribute_schema_id_fkey'),
    sa.PrimaryKeyConstraint('id', name=u'attribute_pkey')
    )
    op.create_table('entity_schema',
    sa.Column('entity_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('schema_id', sa.INTEGER(), autoincrement=False, nullable=True),
    sa.ForeignKeyConstraint(['entity_id'], [u'entity.id'], name=u'entity_schema_entity_id_fkey'),
    sa.ForeignKeyConstraint(['schema_id'], [u'schema.id'], name=u'entity_schema_schema_id_fkey')
    )
    op.create_table('relation',
    sa.Column('created_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('updated_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.VARCHAR(), autoincrement=False, nullable=False),
    sa.Column('schema_id', sa.INTEGER(), autoincrement=False, nullable=True),
    sa.Column('source_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('target_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.ForeignKeyConstraint(['schema_id'], [u'schema.id'], name=u'relation_schema_id_fkey'),
    sa.ForeignKeyConstraint(['source_id'], [u'entity.id'], name=u'relation_source_id_fkey'),
    sa.ForeignKeyConstraint(['target_id'], [u'entity.id'], name=u'relation_target_id_fkey'),
    sa.PrimaryKeyConstraint('id', name=u'relation_pkey')
    )
    op.create_table('property',
    sa.Column('created_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('updated_at', postgresql.TIMESTAMP(), autoincrement=False, nullable=True),
    sa.Column('id', sa.INTEGER(), primary_key=True),
    sa.Column('schema_id', sa.INTEGER(), autoincrement=False, nullable=True),
    sa.Column('name', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('value', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('source_url', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('active', sa.BOOLEAN(), autoincrement=False, nullable=True),
    sa.Column('obj', sa.VARCHAR(length=20), autoincrement=False, nullable=True),
    sa.Column('entity_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('relation_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.ForeignKeyConstraint(['entity_id'], [u'entity.id'], name=u'property_entity_id_fkey'),
    sa.ForeignKeyConstraint(['relation_id'], [u'relation.id'], name=u'property_relation_id_fkey'),
    sa.ForeignKeyConstraint(['schema_id'], [u'schema.id'], name=u'property_schema_id_fkey'),
    sa.PrimaryKeyConstraint('id', name=u'property_pkey')
    )




def downgrade():
    op.drop_table('schema')
    op.drop_table('attribute')
    op.drop_table('entity_schema')
    op.drop_table('property')
    op.drop_table('entity')
    op.drop_table('account')
    op.drop_table('relation')



from StringIO import StringIO

from flask import Blueprint, request
from flask import send_file
from restpager import Pager
from sqlalchemy import or_, and_

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, request_data, get_limit
from grano.model import Project, File, Permission
from grano.logic import files, entities
from grano.logic.references import ProjectRef
from grano.lib.exc import Gone
from grano.core import db
from grano.views.cache import validate_cache
from grano import authz


blueprint = Blueprint('files_api', __name__)


@blueprint.route('/api/1/files', methods=['GET'])
def index():
    query = File.all()
    query = query.join(Project)
    query = query.outerjoin(Permission)
    query = query.filter(or_(Project.private == False, # noqa
        and_(Permission.reader == True, # noqa
             Permission.account == request.account)))

    if request.args.get('project'):
        query = query.filter(Project.slug == request.args.get('project'))

    query = query.distinct()
    pager = Pager(query)
    validate_cache(keys=pager.cache_keys())
    return jsonify(pager, index=True)


@blueprint.route('/api/1/files', methods=['POST', 'PUT'])
def create():
    data = request_data({'author': request.account})
    project = ProjectRef().get(data.get('project'))
    data['project'] = project
    authz.require(authz.project_edit(project))
    file_ = files.save(data, request.files.get('file'))
    db.session.commit()
    return jsonify(file_)


@blueprint.route('/api/1/files/<id>', methods=['GET'])
def view(id):
    file = object_or_404(File.by_id(id))
    authz.require(authz.project_read(file.project))
    return jsonify(file)


@blueprint.route('/api/1/files/<id>/_serve', methods=['GET'])
def serve(id):
    file = object_or_404(File.by_id(id))
    authz.require(authz.project_read(file.project))
    sio = StringIO()
    sio.write(file.data)
    sio.seek(0)
    res = send_file(sio, mimetype=file.mime_type)
    res.headers['Content-Disposition'] = 'filename=%s' % file.file_name
    return res


@blueprint.route('/api/1/files/<id>/_table', methods=['GET'])
def table(id):
    file = object_or_404(File.by_id(id))
    authz.require(authz.project_read(file.project))
    limit = get_limit(10)
    validate_cache(keys={'id': file.id, 'limit': limit})
    return jsonify(files.as_table(file, limit))


@blueprint.route('/api/1/files/<id>', methods=['DELETE'])
def delete(id):
    file_ = object_or_404(File.by_id(id))
    authz.require(authz.project_edit(file_.project))
    entities.delete(file_)
    db.session.commit()
    raise Gone()



import abc


class EntityChangeProcessor(object):
    """ An entity change processor gets notified whenever there is a
    change to an entity so that it can perform related actions.

    This may happen out of band (ie. on a queue or batch job), thus
    changes may not be applied immediately.
    """

    @abc.abstractmethod
    def entity_changed(self, entity_id, operation):
        """ Notify the plugin that an entity has changed. The plugin
        will only receive the ID and must query for the object itself. """


class RelationChangeProcessor(object):
    """ A relation change processor gets notified whenever there is a
    change to a relation so that it can perform related actions.

    This may happen out of band (ie. on a queue or batch job), thus
    changes may not be applied immediately.
    """

    @abc.abstractmethod
    def relation_changed(self, relation_id, operation):
        """ Notify the plugin that a relation has changed. The plugin
        will only receive the ID and must query for the object itself. """


class ProjectChangeProcessor(object):
    """ A project change processor gets notified whenever there is a
    change to a project's settings so that it can perform related actions.

    This may happen out of band (ie. on a queue or batch job), thus
    changes may not be applied immediately.
    """

    @abc.abstractmethod
    def project_changed(self, project_slug, operation):
        """ Notify the plugin that a project has changed. The plugin
        will only receive the ID and must query for the object itself. """


class SchemaChangeProcessor(object):
    """ A schema change processor gets notified whenever there is a
    change to a schema definition so that it can perform related actions.

    This may happen out of band (ie. on a queue or batch job), thus
    changes may not be applied immediately.
    """

    @abc.abstractmethod
    def schema_changed(self, project_slug, schema_name, operation):
        """ Notify the plugin that a project has changed. The plugin
        will only receive the ID and must query for the object itself. """


class Startup(object):
    """ This interface will be called when grano is started and allows
    plugins to register additional functionality such as flask views.
    """

    @abc.abstractmethod
    def configure(self, manager):
        """ Run this on startup. """


class Periodic(object):
    """ This interface will be called periodically, to execute house
    keeping jobs. The exact period is up to the type of deployment.
    """

    @abc.abstractmethod
    def run(self):
        """ Run an action. """



import unittest

from grano.test.util import make_test_app
from grano.ql.query import run as query
from grano.ql.query import PARENT_ID


class TestQuery(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.app = make_test_app()

    def test_single_entity_id(self):
        res = query({'id': None}).to_dict()
        assert 'id' in res, res
        assert res['id'] is not None, res

    def test_list_entity_id(self):
        res = query([{'id': None}]).to_dict()
        assert len(res), res
        assert 'id' in res[0], res
        assert res[0]['id'] is not None, res

    def test_default_fields(self):
        res = query({'*': None}).to_dict()
        assert 'id' in res, res
        assert res['id'] is not None, res
        assert res['status'] is not None, res

    def test_author_nested(self):
        res = query({'author': None}).to_dict()
        assert 'author' in res, res
        assert res['author'] is not None, res
        assert res['author']['login'] is not None, res

    def test_author_filter(self):
        res = query({'author': {'login': '_system'}}).to_dict()
        assert 'author' in res, res
        assert res['author'] is not None, res
        assert res['author']['login'] == '_system', res

    def test_schema_nested(self):
        res = query({'schemata': None}).to_dict()
        assert 'schemata' in res, res
        assert res['schemata'] is not None, res
        assert res['schemata']['name'] is not None, res

    def test_schema_filter(self):
        res = query({'schemata': 'fellow'}).to_dict()
        assert 'schemata' in res, res
        assert res['schemata'] is not None, res
        assert res['schemata']['name'] == 'fellow', res

    def test_project_nested(self):
        res = query({'project': None}).to_dict()
        assert 'project' in res, res
        assert res['project'] is not None, res
        assert res['project']['slug'] is not None, res
        assert PARENT_ID not in res['project'], res

    def test_nested_relation_inbound(self):
        res = query({'inbound': {'id': None}}).to_dict()
        assert 'inbound' in res, res
        assert res['inbound'] is not None, res
        assert res['inbound']['id'] is not None, res

    def test_nested_relation_inbound_source(self):
        res = query({'inbound': {'id': None, 'source': None}}).to_dict()
        assert 'inbound' in res, res
        assert res['inbound'] is not None, res
        assert res['inbound']['id'] is not None, res
        assert res['inbound']['source']['id'] is not None, res

    def test_property_nested(self):
        res = query({'properties': {'name': None}}).to_dict()
        assert 'properties' in res, res
        assert res['properties'] is not None, res
        assert res['properties']['name'] is not None, res
        assert res['properties']['name']['source_url'] is not None, res
        assert res['properties']['name']['value'] is not None, res

    def test_property_filter(self):
        name = 'BBC'
        res = query({'properties': {'name': name}}).to_dict()
        assert 'properties' in res, res
        assert res['properties'] is not None, res
        assert res['properties']['name'] is not None, res
        assert res['properties']['name']['value'] == name, res

    def test_property_filter_none(self):
        name = 'Banana'
        res = query({'properties': {'name': name}}).to_dict()
        assert 'properties' in res, res
        assert res['id'] is None, res
        assert res['properties'] is not None, res
        assert res['properties']['name'] is not None, res

    def test_property_all(self):
        res = query({'properties': {'*': None}}).to_dict()
        assert 'properties' in res, res
        assert res['properties'] is not None, res
        assert '*' not in res['properties'], res
        assert res['properties']['name'] is not None, res

    def test_entity_all(self):
        res = query({'*': None}).to_dict()
        assert 'properties' in res, res
        assert res['properties'] is not None, res
        assert '*' not in res['properties'], res
        assert res['properties']['name'] is not None, res

if __name__ == '__main__':
    unittest.main()



from flask import Blueprint, request
from restpager import Pager

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404
from grano.model import Pipeline, LogEntry
from grano.views.cache import validate_cache
from grano import authz


blueprint = Blueprint('log_entries_api', __name__)


@blueprint.route('/api/1/pipelines/<pipeline_id>/log', methods=['GET'])
def index(pipeline_id):
    pipeline = object_or_404(Pipeline.by_id(pipeline_id))
    authz.require(authz.project_read(pipeline.project))

    query = LogEntry.all()
    query = query.filter(LogEntry.pipeline == pipeline)

    if request.args.get('level'):
        query = query.filter(LogEntry.level == request.args.get('level'))

    pager = Pager(query)
    validate_cache(keys=pager.cache_keys())
    return jsonify(pager, index=True)


@blueprint.route('/api/1/pipelines/<pipeline_id>/log/<id>', methods=['GET'])
def view(pipeline_id, id):
    pipeline = object_or_404(Pipeline.by_id(pipeline_id))
    authz.require(authz.project_read(pipeline.project))
    log_entry = object_or_404(LogEntry.by_id(id))
    return jsonify(log_entry)



from grano.core import db, url_for
from grano.model.common import IntBase


class Permission(db.Model, IntBase):
    __tablename__ = 'grano_permission'

    reader = db.Column(db.Boolean)
    editor = db.Column(db.Boolean)
    admin = db.Column(db.Boolean)

    account_id = db.Column(db.Integer, db.ForeignKey('grano_account.id'))
    project_id = db.Column(db.Integer, db.ForeignKey('grano_project.id'))

    @classmethod
    def by_project_and_id(cls, project, id):
        q = db.session.query(cls).filter_by(id=id)
        q = q.filter_by(project=project)
        return q.first()

    def to_dict_index(self):
        return {
            'id': self.id,
            'reader': self.reader,
            'editor': self.editor,
            'admin': self.admin,
            'project': self.project.to_dict_short(),
            'account': self.account.to_dict_index(),
            'api_url': url_for('permissions_api.view', slug=self.project.slug,
                               id=self.id)
        }

    def to_dict(self):
        return self.to_dict_index()



import logging
from datetime import datetime

from grano.core import db, url_for
from grano.model.util import MutableDict, JSONEncodedDict
from grano.model.common import IntBase


# this is based on OpenSpending, see:
# https://github.com/openspending/openspending/blob/master/openspending/model/run.py


class Pipeline(db.Model, IntBase):
    __tablename__ = 'grano_pipeline'

    STATUS_PENDING = 'pending'
    STATUS_RUNNING = 'running'
    STATUS_FAILED = 'failed'
    STATUS_COMPLETE = 'complete'

    operation = db.Column(db.Unicode)
    label = db.Column(db.Unicode)
    status = db.Column(db.Unicode)
    percent_complete = db.Column(db.Integer, default=int)

    project_id = db.Column(db.Integer, db.ForeignKey('grano_project.id'))
    author_id = db.Column(db.Integer, db.ForeignKey('grano_account.id'))

    config = db.Column(MutableDict.as_mutable(JSONEncodedDict))

    started_at = db.Column(db.DateTime, default=datetime.utcnow)
    ended_at = db.Column(db.DateTime)

    entries = db.relationship('LogEntry', backref='pipeline', lazy='dynamic',
                              cascade='all, delete, delete-orphan')

    def has_errors(self):
        q = self.entries.filter_by(level=logging.ERROR)
        return q.count() > 0

    def to_dict_index(self):
        return {
            'id': self.id,
            'label': self.label,
            'project': self.project.to_dict_short() if self.project else None,
            'author': self.author.to_dict_index(),
            'api_url': url_for('pipelines_api.view', id=self.id),
            'operation': self.operation,
            'status': self.status,
            'created_at': self.created_at,
            'updated_at': self.updated_at,
            'started_at': self.started_at,
            'ended_at': self.ended_at,
            'config': self.config,
            'percent_complete': self.percent_complete
        }

    def to_dict(self):
        """ Full serialization of the file metadata. """
        data = self.to_dict_index()
        return data


class LogEntry(db.Model, IntBase):
    __tablename__ = 'grano_log_entry'

    pipeline_id = db.Column(db.Integer, db.ForeignKey('grano_pipeline.id'))

    level = db.Column(db.Integer)
    message = db.Column(db.Unicode)
    error = db.Column(db.Unicode)

    details = db.Column(MutableDict.as_mutable(JSONEncodedDict))

    def to_dict_index(self):
        return {
            'id': self.id,
            'level': self.level,
            'message': self.message,
            'error': self.error,
            'api_url': url_for('log_entries_api.view_entry',
                               pipeline_id=self.pipeline.id, id=self.id),
            'created_at': self.created_at,
            'updated_at': self.updated_at
        }

    def to_dict(self):
        """ Full serialization of the file metadata. """
        data = self.to_dict_index()
        data['details'] = self.details
        return data



from flask import Blueprint, request
from restpager import Pager
from sqlalchemy.orm import aliased
from sqlalchemy import not_

from grano.lib.serialisation import jsonify
from grano.lib.exc import BadRequest, Gone
from grano.lib.args import object_or_404, request_data
from grano.model import Entity, Property, Project
from grano.logic import entities
from grano.logic.references import ProjectRef
from grano.core import db, url_for
from grano.views import filters, facets, sorters
from grano.views.cache import validate_cache
from grano import authz


blueprint = Blueprint('entities_api', __name__)


@blueprint.route('/api/1/entities', methods=['GET'])
def index():
    alias = aliased(Entity)
    query = db.session.query(alias)
    query = filters.for_entities(query, alias)
    query = sorters.for_entities(query, alias)
    pager = Pager(query)
    validate_cache(keys=pager.cache_keys())
    result = pager.to_dict()
    result['facets'] = facets.for_entities()
    return jsonify(result, index=True)


@blueprint.route('/api/1/entities', methods=['POST', 'PUT'])
def create():
    data = request_data({'author': request.account})
    project = ProjectRef().get(data.get('project'))
    data['project'] = project
    authz.require(authz.project_edit(project))
    entity = entities.save(data, files=request.files)
    db.session.commit()
    return jsonify(entity)


@blueprint.route('/api/1/entities/<id>', methods=['GET'])
def view(id):
    entity = object_or_404(Entity.by_id(id))
    authz.require(authz.entity_read(entity))
    return jsonify(entity)


@blueprint.route('/api/1/entities/_suggest', methods=['GET'])
def suggest():
    if 'q' not in request.args or not len(request.args.get('q').strip()):
        raise BadRequest("Missing the query ('q' parameter).")

    q = db.session.query(Property)
    q = q.join(Entity)
    q = q.filter(Entity.project_id.in_(authz.permissions().get('reader')))
    q = q.filter(Property.name == 'name')
    q = q.filter(Property.active == True) # noqa
    q = q.filter(Property.entity_id != None) # noqa
    q = q.filter(Property.value_string.ilike(request.args.get('q') + '%'))
    if 'project' in request.args:
        q = q.join(Project)
        q = q.filter(Project.slug == request.args.get('project'))
    if 'exclude' in request.args:
        ents = request.args.getlist('exclude')
        q = q.filter(not_(Property.entity_id.in_(ents)))
    q = q.distinct()
    pager = Pager(q)

    data = []

    def convert(props):
        for prop in props:
            data.append({
                'properties': {
                    'name': prop.to_dict_index(),
                },
                'id': prop.entity_id,
                'api_url': url_for('entities_api.view', id=prop.entity_id)
            })
        return data

    validate_cache(keys='#'.join([d['name'] for d in data]))
    return jsonify(pager.to_dict(results_converter=convert))


@blueprint.route('/api/1/entities/<id>', methods=['POST', 'PUT'])
def update(id):
    entity = object_or_404(Entity.by_id(id))
    authz.require(authz.entity_edit(entity))
    data = request_data({'author': request.account})
    entity = entities.save(data, files=request.files, entity=entity)
    db.session.commit()
    return jsonify(entity)


@blueprint.route('/api/1/entities/_merge', methods=['POST', 'PUT'])
def merge():
    validator = entities.MergeValidator()
    data = validator.deserialize(request_data())
    authz.require(authz.project_edit(data['orig'].project))

    if data['orig'].id == data['dest'].id:
        raise BadRequest('Origin and destination are identical.')

    if data['orig'].project_id != data['dest'].project_id:
        raise BadRequest('Entities belong to different projects.')

    dest = entities.merge(data['orig'], data['dest'])
    db.session.commit()
    return jsonify(dest)


@blueprint.route('/api/1/entities/<id>', methods=['DELETE'])
def delete(id):
    entity = object_or_404(Entity.by_id(id))
    authz.require(authz.entity_edit(entity))
    entities.delete(entity)
    db.session.commit()
    raise Gone()



"""add file relation

Revision ID: 52f4346513a8
Revises: d7067cd4169
Create Date: 2014-04-02 16:33:16.710455

"""

# revision identifiers, used by Alembic.
revision = '52f4346513a8'
down_revision = 'd7067cd4169'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('grano_file',
        sa.Column('created_at', sa.DateTime(), nullable=True),
        sa.Column('updated_at', sa.DateTime(), nullable=True),
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('file_name', sa.Unicode(), nullable=True),
        sa.Column('mime_type', sa.Unicode(), nullable=True),
        sa.Column('project_id', sa.Integer(), nullable=True),
        sa.Column('author_id', sa.Integer(), nullable=True),
        sa.Column('data', sa.LargeBinary(), nullable=True),
        sa.ForeignKeyConstraint(['author_id'], ['grano_account.id'], ),
        sa.ForeignKeyConstraint(['project_id'], ['grano_project.id'], ),
        sa.PrimaryKeyConstraint('id')
    )
    #op.drop_table('role')


def downgrade():
    op.drop_table('grano_file')



from grano.core import db, url_for
from grano.model.common import IntBase
from grano.model.util import MutableDict, JSONEncodedDict
from grano.model.attribute import Attribute


ENTITY_DEFAULT = 'Entity'
RELATION_DEFAULT = 'Relation'


class Schema(db.Model, IntBase):
    __tablename__ = 'grano_schema'

    name = db.Column(db.Unicode())
    label = db.Column(db.Unicode())
    hidden = db.Column(db.Boolean())
    obj = db.Column(db.Unicode())
    meta = db.Column(MutableDict.as_mutable(JSONEncodedDict))
    project_id = db.Column(db.Integer, db.ForeignKey('grano_project.id'))
    parent_id = db.Column(db.Integer, db.ForeignKey('grano_schema.id'),
                          nullable=True)

    local_attributes = db.relationship(Attribute, backref='schema',
                                       lazy='dynamic',
                                       cascade='all, delete, delete-orphan')
    relations = db.relationship('Relation', backref='schema', lazy='dynamic',
                                cascade='all, delete, delete-orphan')
    entities = db.relationship('Entity', backref='schema', lazy='dynamic',
                               cascade='all, delete, delete-orphan')
    children = db.relationship('Schema', lazy='dynamic',
                               backref=db.backref('parent',
                                                  remote_side='Schema.id'))

    @property
    def inherited_attributes(self):
        if self.parent is None:
            return []
        return self.parent.attributes

    @property
    def attributes(self):
        return list(self.local_attributes) + self.inherited_attributes

    def get_attribute(self, name):
        for attribute in self.attributes:
            if attribute.name == name:
                return attribute

    def is_circular(self, path=None):
        if path is None:
            path = []
        if self.name in path:
            return True
        elif self.parent is None:
            return False
        else:
            path.append(self.name)
            return self.parent.is_circular(path)

    def is_parent(self, other):
        if self.parent is None:
            return False
        if self.parent == other:
            return True
        return self.parent.is_parent(other)

    def common_parent(self, other):
        if self == other or self.is_parent(other):
            return self
        return self.common_parent(other.parent)

    @classmethod
    def by_name(cls, project, name):
        q = db.session.query(cls).filter_by(name=name)
        q = q.filter_by(project=project)
        return q.first()

    @classmethod
    def by_obj_name(cls, project, obj, name):
        q = db.session.query(cls)
        q = q.filter_by(project=project)
        q = q.filter_by(name=name)
        q = q.filter_by(obj=obj)
        return q.first()

    def to_dict_index(self):
        return {
            'name': self.name,
            'label': self.label,
            'hidden': self.hidden,
            'obj': self.obj,
            'api_url': url_for('schemata_api.view',
                               slug=self.project.slug,
                               name=self.name)
        }

    def to_dict(self):
        data = self.to_dict_index()
        data['id'] = self.id
        data['meta'] = self.meta
        if self.parent is not None:
            data['parent'] = self.parent.to_dict_index()
        else:
            data['parent'] = None
        data['project'] = self.project.to_dict_short()
        data['attributes'] = [a.to_dict() for a in self.local_attributes]
        for attr in self.inherited_attributes:
            d = attr.to_dict()
            d['inherited'] = True
            data['attributes'].append(d)
        return data



import colander

from grano.lib.exc import BadRequest
from grano.model import Project, Entity, Account
from grano.model import Schema, File


class Ref(object):
    def serialize(self, node, appstruct):
        if appstruct is colander.null:
            return colander.null
        return None

    def deserialize(self, node, cstruct):
        if cstruct is colander.null:
            return colander.null
        value = self.decode(node, cstruct)
        if value is None:
            raise colander.Invalid(node, 'Missing')
        return value

    def cstruct_children(self, node, cstruct):
        return []

    def get(self, cstruct):
        if cstruct is None:
            raise BadRequest()
        project = self.decode(None, cstruct)
        if project is None:
            raise BadRequest()
        return project


class ProjectRef(Ref):

    def decode(self, node, cstruct):
        if isinstance(cstruct, Project):
            return cstruct
        if isinstance(cstruct, basestring):
            return Project.by_slug(cstruct)
        if isinstance(cstruct, dict):
            if cstruct.get('slug'):
                return Project.by_slug(cstruct.get('slug'))
        return None


class EntityRef(Ref):

    def __init__(self, project=None):
        self.project = project

    def decode(self, node, cstruct):
        if isinstance(cstruct, Entity):
            if cstruct.project == self.project:
                return cstruct
        if isinstance(cstruct, basestring):
            entity = Entity.by_id(cstruct)
            if entity is None:
                return None
            if entity.project == self.project:
                return entity
        if isinstance(cstruct, dict):
            if cstruct.get('id'):
                entity = Entity.by_id(cstruct.get('id'))
                if self.project is None:
                    return entity
                if entity is not None and entity.project == self.project:
                    return entity
        return None


class AccountRef(Ref):

    def decode(self, node, cstruct):
        if isinstance(cstruct, Account):
            return cstruct
        if isinstance(cstruct, int):
            return Account.by_id(cstruct)
        if isinstance(cstruct, dict):
            if cstruct.get('id'):
                return Account.by_id(cstruct.get('id'))
        return None


class FileRef(Ref):

    def decode(self, node, cstruct):
        if isinstance(cstruct, File):
            return cstruct
        if isinstance(cstruct, (int, basestring)):
            return File.by_id(cstruct)
        if isinstance(cstruct, dict):
            if cstruct.get('id'):
                return File.by_id(cstruct.get('id'))
        return None


class SchemaRef(Ref):

    def __init__(self, project):
        self.project = project

    def decode(self, node, cstruct):
        if isinstance(cstruct, Schema):
            return cstruct
        if isinstance(cstruct, basestring):
            return Schema.by_name(self.project, cstruct)
        if isinstance(cstruct, dict):
            if cstruct.get('name'):
                return Schema.by_name(self.project,
                                      cstruct.get('name'))
        return None



import os
import logging
from unicodecsv import DictReader, DictWriter
from sqlalchemy.orm import aliased

from grano.core import db
from grano.logic import entities, pipelines, imports
from grano.model import Entity, Property, Schema

log = logging.getLogger(__name__)


## Import commands
def import_aliases(project, author, path):
    """ Set up a data pipeline and execute it. """
    with open(path, 'r') as fh:
        config = {'mapping': {
            'canonical': {'attribute': 'canonical'},
            'alias': {'attribute': 'alias'}
        }}
        name = os.path.basename(path)
        pipeline = pipelines.create(project, 'import',
                                    name, config, author)
        pipelines.start(pipeline)
        imports.import_aliases(pipeline, fh)
        pipelines.finish(pipeline)


## Export commands
def export_aliases(project, fh):
    """ Dump a list of all entity names to a CSV file. The table will contain
    the active name of each entity, and one of the other existing names as an
    alias. """

    writer = DictWriter(fh, ['entity_id', 'schema', 'alias', 'canonical'])
    writer.writeheader()

    alias = aliased(Property)
    canonical = aliased(Property)
    schema = aliased(Schema)
    q = db.session.query(alias.value_string.label('alias'), alias.entity_id)
    q = q.join(Entity)
    q = q.join(schema)
    q = q.join(canonical)
    q = q.filter(Entity.project_id == project.id)
    q = q.filter(alias.entity_id != None) # noqa
    q = q.filter(alias.name == 'name')
    q = q.filter(canonical.name == 'name')
    q = q.filter(canonical.active == True) # noqa
    q = q.add_columns(canonical.value_string.label('canonical'))
    q = q.add_columns(schema.name.label('schema'))
    for row in q.all():
        writer.writerow({
            'entity_id': str(row.entity_id),
            'schema': row.schema,
            'alias': row.alias,
            'canonical': row.canonical
        })



"""foreign keys for project

Revision ID: 2b65d3dfa1de
Revises: 40941ab93ba
Create Date: 2014-02-16 13:19:42.398726

"""

# revision identifiers, used by Alembic.
revision = '2b65d3dfa1de'
down_revision = '40941ab93ba'

import uuid
from alembic import op
import sqlalchemy as sa
from sqlalchemy.sql import table, column


def upgrade():
    account = table('account',
        column('login', sa.String),
        column('api_key', sa.String)
    )
    engine = op.get_bind()

    op.execute(account.insert().values({'login': '_system', 'api_key': uuid.uuid4().hex}))

    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('entity', sa.Column('author_id', sa.Integer(), nullable=True))
    op.add_column('entity', sa.Column('project_id', sa.Integer(), nullable=True))
    op.add_column('project', sa.Column('author_id', sa.Integer(), nullable=True))
    
    if engine.dialect.name != 'sqlite':
        op.drop_column('project', 'owner_id')
    
    op.add_column('property', sa.Column('author_id', sa.Integer(), nullable=True))
    op.add_column('relation', sa.Column('author_id', sa.Integer(), nullable=True))
    op.add_column('relation', sa.Column('project_id', sa.Integer(), nullable=True))
    op.add_column('schema', sa.Column('project_id', sa.Integer(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('schema', 'project_id')
    op.drop_column('relation', 'project_id')
    op.drop_column('relation', 'author_id')
    op.drop_column('property', 'author_id')
    op.add_column('project', sa.Column('owner_id', sa.INTEGER(), nullable=True))
    op.drop_column('project', 'author_id')
    op.drop_column('entity', 'project_id')
    op.drop_column('entity', 'author_id')
    ### end Alembic commands ###



import json

from flask import request
from restpager.args import arg_bool, arg_int, get_limit, get_offset # noqa
from grano.lib.exc import BadRequest, NotFound


def request_data(overlay={}):
    """ Decode a JSON-formatted POST body. """
    data = request.json
    if data is None:
        try:
            data = json.loads(request.form.get('data'))
        except (ValueError, TypeError):
            data = dict(request.form.items())
    data.update(overlay)
    return data


def object_or_404(obj):
    if obj is None:
        raise NotFound()
    return obj


def single_arg(name, default=None):
    vals = [v for v in request.args.getlist(name) if v.strip()]
    if len(vals) == 0:
        return default
    elif len(vals) > 1:
        raise BadRequest('Too many values given for: %s' % name)
    else:
        return vals[0]



from flask import Blueprint, request
from restpager import Pager
from sqlalchemy import or_, and_

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404
from grano.model import Project, Permission, Pipeline
from grano.views.cache import validate_cache
from grano import authz


blueprint = Blueprint('pipelines_api', __name__)


@blueprint.route('/api/1/pipelines', methods=['GET'])
def index():
    query = Pipeline.all()
    query = query.join(Project)
    query = query.outerjoin(Permission)
    query = query.filter(or_(Project.private==False,
        and_(Permission.reader==True, Permission.account==request.account)))

    if request.args.get('project'):
        query = query.filter(Project.slug==request.args.get('project'))

    if request.args.get('operation'):
        query = query.filter(Pipeline.operation == request.args.get('operation'))

    query = query.order_by(Pipeline.updated_at.desc())
    query = query.distinct()

    pager = Pager(query)
    validate_cache(keys=pager.cache_keys())
    return jsonify(pager, index=True)


@blueprint.route('/api/1/pipelines/<id>', methods=['GET'])
def view(id):
    pipeline = object_or_404(Pipeline.by_id(id))
    authz.require(authz.project_read(pipeline.project))
    return jsonify(pipeline)



from flask import Blueprint
from restpager import Pager

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, request_data, arg_bool
from grano.model import Schema, Project
from grano.logic import schemata
from grano.lib.exc import Gone
from grano.core import db
from grano.views.cache import validate_cache
from grano import authz


blueprint = Blueprint('schemata_api', __name__)


@blueprint.route('/api/1/projects/<slug>/schemata', methods=['GET'])
def index(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))
    validate_cache(last_modified=project.updated_at)
    query = Schema.all()
    query = query.filter_by(project=project)
    pager = Pager(query, slug=slug)
    return jsonify(pager, index=not arg_bool('full'))


@blueprint.route('/api/1/projects/<slug>/schemata', methods=['POST', 'PUT'])
def create(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    data = request_data({'project': project})
    schema = schemata.save(data)
    db.session.commit()
    return jsonify(schema, status=201)


@blueprint.route('/api/1/projects/<slug>/schemata/<name>', methods=['GET'])
def view(slug, name):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))
    if not project.private:
        validate_cache(last_modified=project.updated_at)
    schema = object_or_404(Schema.by_name(project, name))
    return jsonify(schema)


@blueprint.route('/api/1/projects/<slug>/schemata/<name>',
                 methods=['POST', 'PUT'])
def update(slug, name):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    schema = object_or_404(Schema.by_name(project, name))
    data = request_data({'project': project})
    schema = schemata.save(data, schema=schema)
    db.session.commit()
    return jsonify(schema)


@blueprint.route('/api/1/projects/<slug>/schemata/<name>', methods=['DELETE'])
def delete(slug, name):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    schema = object_or_404(Schema.by_name(project, name))
    deleted = schemata.delete(schema)
    db.session.commit()
    if deleted:
        raise Gone()
    else:
        return jsonify(schema)



import unittest
import flask
from grano import authz
from grano import core
from grano.core import db
from grano.manage import app
from grano.model import Project, Permission, Entity, Relation
from grano.test.fixtures import create_fixtures


def make_test_app(use_cookies=False):
    """For some reason this does not work when imported from util.
    Error is:
        AttributeError: 'FlaskClient' object has no attribute 'test_request_context'
    """
    app.config['TESTING'] = True
    app.config['CACHE'] = False
    app.config['PLUGINS'] = []
    app.config['CELERY_ALWAYS_EAGER'] = True
    app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///:memory:'
    db.create_all()
    create_fixtures()
    return app


class BaseAuthTestCase(unittest.TestCase):
    def login(self, username, password):
        return self.app.post('/login', data=dict(
            username=username,
            password=password
        ), follow_redirects=True)
    
    def logout(self):
        return self.app.get('/logout', follow_redirects=True)


def _project_and_permission(
        private=False,
        reader=False,
        editor=False,
        admin=False, account_id=1):
    project = Project(private=private)
    db.session.add(project)
    db.session.commit()
    perm = Permission(account_id=account_id, project_id=project.id,
        reader=reader, editor=editor, admin=admin)
    db.session.add(perm)
    db.session.commit()
    return project, perm


class ProjectAuthTestCase(BaseAuthTestCase):

    def setUp(self):
        self.app = make_test_app()

    def test_public_project_read(self):
        project, permission = _project_and_permission()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.project_read(project))

    def test_private_project_read__no_perm(self):
        project, permission = _project_and_permission(private=True)
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.project_read(project))
      
    def test_private_project_read__admin_only(self):
        """Perms are designed such that admin/editor cannot read private
        projects unless they have explicit read permission."""
        project, permission = _project_and_permission(
            private=True, admin=True)
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.project_read(project))
      
    def test_private_project_read__editor_only(self):
        project, permission = _project_and_permission(
            private=True, editor=True)
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.project_read(project))
      
    def test_private_project_read__reader_only(self):
        with self.app.test_request_context():
            project, permission = _project_and_permission(
                private=True, reader=True)
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.project_read(project))
      
        
class EntityAuthTestCase(BaseAuthTestCase):

    def setUp(self):
        self.app = make_test_app()

    def test_public_project_entity_read__default_status(self):
        project, permission = _project_and_permission()
        entity = Entity(project=project)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.entity_read(entity))

    def test_public_project_entity_read__publish_status(self):
        project, permission = _project_and_permission()
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.entity_read(entity))

    def test_public_project_entity_read__draft_status(self):
        project, permission = _project_and_permission()
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.entity_read(entity))

    def test_private_project_entity_read__publish_status_reader(self):
        project, permission = _project_and_permission(private=True,
            reader=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.entity_read(entity))

    def test_private_project_entity_read__draft_status_reader(self):
        project, permission = _project_and_permission(private=True,
            reader=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.entity_read(entity))

    def test_private_project_entity_read__draft_status_editor(self):
        project, permission = _project_and_permission(private=True,
            editor=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.entity_read(entity))

    def test_private_project_entity_read__draft_status_admin(self):
        project, permission = _project_and_permission(private=True,
            admin=True)
        entity = Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        db.session.add(entity)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.entity_read(entity))

class RelationAuthTestCase(BaseAuthTestCase):

    def setUp(self):
        self.app = make_test_app()

    def test_entity_read__unreadable(self):
        project, permission = _project_and_permission()
        entity_source = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        entity_target = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD-1)
        db.session.add(entity_source)
        db.session.add(entity_target)
        rel = Relation(source=entity_source, target=entity_target)
        db.session.add(rel)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.relation_read(rel))

    def test_entity_read__readable(self):
        project, permission = _project_and_permission()
        entity_source = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        entity_target = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity_source)
        db.session.add(entity_target)
        rel = Relation(source=entity_source, target=entity_target)
        db.session.add(rel)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.relation_read(rel))

    def test_entity_edit__editable(self):
        project, permission = _project_and_permission(editor=True)
        entity_source = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        entity_target = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity_source)
        db.session.add(entity_target)
        rel = Relation(source=entity_source, target=entity_target)
        db.session.add(rel)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.relation_edit(rel))

    def test_entity_edit__uneditable(self):
        project, permission = _project_and_permission(editor=False)
        entity_source = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        entity_target = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity_source)
        db.session.add(entity_target)
        rel = Relation(source=entity_source, target=entity_target)
        db.session.add(rel)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.relation_edit(rel))

    def test_entity_manage__manageable(self):
        project, permission = _project_and_permission(admin=True)
        entity_source = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        entity_target = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity_source)
        db.session.add(entity_target)
        rel = Relation(source=entity_source, target=entity_target)
        db.session.add(rel)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertTrue(authz.relation_manage(rel))

    def test_entity_manage__unmanageable(self):
        project, permission = _project_and_permission(admin=False)
        entity_source = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        entity_target = \
            Entity(project=project, status=authz.PUBLISHED_THRESHOLD)
        db.session.add(entity_source)
        db.session.add(entity_target)
        rel = Relation(source=entity_source, target=entity_target)
        db.session.add(rel)
        db.session.commit()
        with self.app.test_request_context():
            flask.session['id'] = 1
            self.app.preprocess_request()
            self.assertFalse(authz.relation_manage(rel))


if __name__ == '__main__':
    unittest.main()



from datetime import datetime

from flask import Blueprint, request
from flask import redirect, make_response

from grano.lib.serialisation import jsonify
from grano.core import app, url_for, app_name, app_version
from grano.views.cache import validate_cache, disable_cache
from grano.background import ping


blueprint = Blueprint('base_api', __name__)
startup_time = datetime.utcnow()
ROBOTS = """
User-agent: *
Sitemap: /static/sitemap.xml
Disallow: /harming/humans
Disallow: /ignoring/human/orders
Disallow: /harm/to/self
"""


@blueprint.route('/robots.txt', methods=['GET'])
def robots_txt():
    validate_cache(last_modified=startup_time)
    res = make_response(ROBOTS)
    res.headers['Content-Type'] = 'text/plain'
    return res


@blueprint.route('/favicon.ico', methods=['GET'])
def favicon_ico():
    validate_cache(last_modified=startup_time)
    ico_url = app.config.get('FAVICON_URL',
        'http://assets.pudo.org/img/favicon.ico')
    return redirect(ico_url)


@blueprint.route('/api', methods=['GET'])
@blueprint.route('/api/1', methods=['GET'])
def status():
    validate_cache(last_modified=startup_time)
    return jsonify({
        'service': app_name,
        'status': 'ok',
        'version': app_version,
        'docs': 'http://granoproject.org/docs',
        'api_url': url_for('base_api.status'),
        'services': {
            'projects_index_url': url_for('projects_api.index'),
            'entities_index_url': url_for('entities_api.index'),
            'relations_index_url': url_for('relations_api.index'),
            'sessions_status_url': url_for('sessions_api.status')
        }
    })


@blueprint.route('/api/1/ping', methods=['GET'])
def queue_ping():
    disable_cache()
    ret = ping.delay(message=request.args.get('message'))
    return jsonify({'status': 'sent', 'task': ret.task_name, 'id': ret.id})



from flask import request, session

from grano.core import app
from grano.model import Account
from grano.lib.exc import Unauthorized
from grano.lib.args import arg_bool


@app.before_request
def check_auth():
    api_key = request.headers.get('X-Grano-API-Key') \
        or request.args.get('api_key')
    if session.get('id'):
        request.account = Account.by_id(session.get('id'))
        if request.account is None:
            del session['id']
            raise Unauthorized()
    elif api_key is not None:
        request.account = Account.by_api_key(api_key)
        if request.account is None:
            raise Unauthorized()
        if arg_bool('api_key_cookie'):
            session['id'] = request.account.id
    else:
        request.account = None



import unittest
import json

from util import make_test_app, tear_down_test_app
#from util import create_fixture_user, AUTHZ

class SessionsTestCase(unittest.TestCase):

    def setUp(self):
        self.app = make_test_app()
        #create_fixture_user(self.app)

    def tearDown(self):
        tear_down_test_app()

    def test_sessions_get(self):
        res = self.app.get('/api/1/sessions')
        assert res.status_code == 200, 'OK'
        data = json.loads(res.data)
        assert data.get('logged_in') is False, 'Not logged in'
        

if __name__ == '__main__':
    unittest.main()




from grano.core import db, url_for
from grano.model.common import UUIDBase
from grano.model.property import Property, PropertyBase


class Relation(db.Model, UUIDBase, PropertyBase):
    __tablename__ = 'grano_relation'
    #PropertyClass = RelationProperty

    schema_id = db.Column(db.Integer, db.ForeignKey('grano_schema.id'),
                          index=True)
    source_id = db.Column(db.Unicode, db.ForeignKey('grano_entity.id'),
                          index=True)
    target_id = db.Column(db.Unicode, db.ForeignKey('grano_entity.id'),
                          index=True)
    project_id = db.Column(db.Integer, db.ForeignKey('grano_project.id'))
    author_id = db.Column(db.Integer, db.ForeignKey('grano_account.id'))

    properties = db.relationship(Property,
                                 order_by=Property.created_at.desc(),
                                 cascade='all, delete, delete-orphan',
                                 backref='relation', lazy='dynamic')

    def to_dict_base(self):
        return {
            'id': self.id,
            'properties': {},
            'project': self.project.to_dict_short(),
            'api_url': url_for('relations_api.view', id=self.id),
            'schema': self.schema.to_dict_index(),
            'source': self.source.to_dict_index(),
            'target': self.target.to_dict_index()
        }

    def to_dict(self):
        data = self.to_dict_base()
        for prop in self.active_properties:
            name, prop = prop.to_dict_kv()
            data['properties'][name] = prop
        return data

    def to_dict_index(self):
        data = self.to_dict_base()
        for prop in self.active_properties:
            name, prop = prop.to_dict_kv()
            data['properties'][name] = prop
        return data


class BidiRelation(db.Model):
    __tablename__ = 'grano_relation_bidi'

    id = db.Column(db.Unicode, primary_key=True)
    created_at = db.Column(db.DateTime)
    updated_at = db.Column(db.DateTime)
    reverse = db.Column(db.Boolean)
    
    relation_id = db.Column(db.Unicode)
    source_id = db.Column(db.Unicode)
    target_id = db.Column(db.Unicode)
    project_id = db.Column(db.Integer)
    schema_id = db.Column(db.Integer)
    author_id = db.Column(db.Integer)



from StringIO import StringIO

from grano.core import db, url_for
from grano.model.common import IntBase


class File(db.Model, IntBase):
    __tablename__ = 'grano_file'

    file_name = db.Column(db.Unicode)
    mime_type = db.Column(db.Unicode)

    project_id = db.Column(db.Integer, db.ForeignKey('grano_project.id'))
    author_id = db.Column(db.Integer, db.ForeignKey('grano_account.id'))

    data = db.Column(db.LargeBinary)

    properties = db.relationship('Property', backref='value_file',
        cascade='all, delete, delete-orphan', lazy='dynamic')

    @property
    def fh(self):
        return StringIO(self.data)


    def to_dict_index(self):
        return {
            'id': self.id,
            'project': self.project.to_dict_index(),
            'api_url': url_for('files_api.view', id=self.id),
            'serve_api_url': url_for('files_api.serve', id=self.id),
            'file_name': self.file_name,
            'mime_type': self.mime_type
        }


    def to_dict(self):
        """ Full serialization of the file metadata. """
        return self.to_dict_index()



"""add schema inheritance

Revision ID: 3f0a4c96f763
Revises: 38c7982f4160
Create Date: 2014-09-14 21:47:02.098699

"""

# revision identifiers, used by Alembic.
revision = '3f0a4c96f763'
down_revision = '38c7982f4160'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column(u'grano_attribute', sa.Column('inherited', sa.Boolean(), nullable=True))
    op.add_column(u'grano_schema', sa.Column('parent_id', sa.Integer(), nullable=True))
    

def downgrade():
    op.drop_column(u'grano_schema', 'parent_id')
    op.drop_column(u'grano_attribute', 'inherited')



import colander

from grano.core import db
from grano.model import Permission
from grano.logic.references import ProjectRef, AccountRef


class PermissionValidator(colander.MappingSchema):
    project = colander.SchemaNode(ProjectRef())
    account = colander.SchemaNode(AccountRef())
    reader = colander.SchemaNode(colander.Boolean(), missing=False)
    editor = colander.SchemaNode(colander.Boolean(), missing=False)
    admin = colander.SchemaNode(colander.Boolean(), missing=False)


def save(data, permission=None):
    validator = PermissionValidator()
    data = validator.deserialize(data)

    if permission is None:
        q = Permission.all()
        q = q.filter(Permission.project == data['project'])
        q = q.filter(Permission.account == data['account'])
        permission = q.first()

    if permission is None:
        permission = Permission()
        permission.project = data.get('project')
        permission.account = data.get('account')

    permission.reader = data['reader'] or data['editor'] or data['admin']
    permission.editor = data['editor'] or data['admin']
    permission.admin = data['admin']

    db.session.add(permission)
    db.session.flush()
    return permission


def delete(permission):
    db.session.delete(permission)
    db.session.flush()



"""single schema entity

Revision ID: 1b1b7e728dc4
Revises: 3f0a4c96f763
Create Date: 2014-09-16 10:15:16.427774

"""

# revision identifiers, used by Alembic.
revision = '1b1b7e728dc4'
down_revision = '3f0a4c96f763'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    op.drop_table('grano_entity_schema')
    op.add_column('grano_entity', sa.Column('schema_id', sa.Integer(), nullable=True))
    op.create_index('ix_grano_entity_schema_id', 'grano_entity', ['schema_id'], unique=False)
    op.drop_column('grano_property', 'schema_id')
    

def downgrade():
    op.add_column('grano_property', sa.Column('schema_id', sa.INTEGER(), nullable=True))
    op.drop_index('ix_grano_entity_schema_id', table_name='grano_entity')
    op.drop_column('grano_entity', 'schema_id')
    
    op.create_table('grano_entity_schema',
    sa.Column('entity_id', sa.VARCHAR(), autoincrement=False, nullable=True),
    sa.Column('schema_id', sa.INTEGER(), autoincrement=False, nullable=True),
    sa.ForeignKeyConstraint(['entity_id'], [u'grano_entity.id'], name=u'entity_schema_entity_id_fkey'),
    sa.ForeignKeyConstraint(['schema_id'], [u'grano_schema.id'], name=u'entity_schema_schema_id_fkey')
    )
    



from grano.core import db
from grano.interface import Periodic


COMPUTE_DEGREES = """
UPDATE grano_entity SET degree_in = 0;

UPDATE grano_entity SET degree_in = sq.cnt
    FROM
        (SELECT target_id, COUNT(r.id) AS cnt FROM grano_relation r
            GROUP BY r.target_id) AS sq
    WHERE grano_entity.id = sq.target_id;

UPDATE grano_entity SET degree_out = 0;

UPDATE grano_entity SET degree_out = sq.cnt
    FROM
        (SELECT source_id, COUNT(r.id) AS cnt FROM grano_relation r
            GROUP BY r.source_id) AS sq
    WHERE grano_entity.id = sq.source_id;

UPDATE grano_entity SET degree = degree_in + degree_out;
"""


class Degrees(Periodic):

    def run(self):
        db.session.execute(COMPUTE_DEGREES)
        db.session.commit()



import json
import logging

from sqlalchemy import or_
from flask import Blueprint, request, url_for

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, get_limit
from grano.core import db, app_name
from grano.lib.exc import BadRequest
from grano.model import Entity, Project, Property
from grano.model import Schema, Attribute
from grano import authz

from grano.logic.reconcile import find_matches


blueprint = Blueprint('reconcile_api', __name__)
log = logging.getLogger(__name__)


def reconcile_index(project):
    domain = url_for('base_api.status', _external=True).strip('/')
    urlp = domain + '/entities/{{id}}'
    meta = {
        'name': '%s: %s' % (app_name, project.label),
        'identifierSpace': 'http://rdf.freebase.com/ns/type.object.id',
        'schemaSpace': 'http://rdf.freebase.com/ns/type.object.id',
        'view': {'url': urlp},
        'preview': {
            'url': urlp + '?preview=true',
            'width': 600,
            'height': 300
        },
        'suggest': {
            'entity': {
                'service_url': domain,
                'service_path': '/projects/' + project.slug + '/suggest_entity'
            },
            'type': {
                'service_url': domain,
                'service_path': '/projects/' + project.slug + '/suggest_type'
            },
            'property': {
                'service_url': domain,
                'service_path': '/projects/' + project.slug + '/suggest_property'
            }
        },
        'defaultTypes': []
    }

    for schema in project.schemata:
        if schema.hidden or schema.obj != 'entity':
            continue
        data = {
            'id': '/%s/%s' % (project.slug, schema.name),
            'name': schema.label
        }
        meta['defaultTypes'].append(data)

    return jsonify(meta)


def reconcile_op(project, query):
    log.info("Reconciling in %s: %r", project.slug, query)
    
    schemata = []
    if 'type' in query:
        schemata = query.get('type')
        if isinstance(schemata, basestring):
            schemata = [schemata]
        schemata = [s.rsplit('/', 1)[-1] for s in schemata]

    properties = []
    if 'properties' in query:
        for p in query.get('properties'):
            properties.append((p.get('pid'), p.get('v')))

    matches = find_matches(project, request.account,
                           query.get('query', ''),
                           schemata=schemata,
                           properties=properties)
    matches = matches.limit(get_limit(default=5))

    results = []
    for match in matches:
        data = {
            'name': match['entity']['name'].value,
            'score': match['score'],
            'type': [{
                'id': '/' + project.slug + '/' + match['entity'].schema.name,
                'name': match['entity'].schema.label
            }],
            'id': match['entity'].id,
            'uri': url_for('entities_api.view', id=match['entity'].id,
                           _external=True),
            'match': False #match['score'] == 100
        }
        results.append(data)

    return {
        'result': results,
        'num': len(results)
    }


@blueprint.route('/api/1/projects/<slug>/reconcile', methods=['GET', 'POST'])
def reconcile(slug):
    """
    Reconciliation API, emulates Google Refine API. See:
    http://code.google.com/p/google-refine/wiki/ReconciliationServiceApi
    """
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))

    # TODO: Add proper support for types and namespacing.
    data = request.args.copy()
    data.update(request.form.copy())

    if 'query' in data:
        # single
        q = data.get('query')
        if q.startswith('{'):
            try:
                q = json.loads(q)
            except ValueError:
                raise BadRequest()
        else:
            q = data
        return jsonify(reconcile_op(project, q))
    elif 'queries' in data:
        # multiple requests in one query
        qs = data.get('queries')
        try:
            qs = json.loads(qs)
        except ValueError:
            raise BadRequest()
        queries = {}
        for k, q in qs.items():
            queries[k] = reconcile_op(project, q)
        return jsonify(queries)
    else:
        return reconcile_index(project)


@blueprint.route('/api/1/projects/<slug>/suggest_entity', methods=['GET', 'POST'])
def suggest_entity(slug):
    """
    Suggest API, emulates Google Refine API. See:
    https://github.com/OpenRefine/OpenRefine/wiki/Reconciliation-Service-API
    """

    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))

    prefix = '%%%s%%' % request.args.get('prefix', '')
    log.info("Suggesting entities in %s: %r", project.slug, prefix)

    q = db.session.query(Entity)
    q = q.join(Property)
    q = q.join(Project)
    q = q.filter(Property.name == 'name')
    q = q.filter(Property.active == True) # noqa
    q = q.filter(Property.entity_id == Entity.id)
    q = q.filter(Property.value_string.ilike(prefix))
    q = q.filter(Project.slug == slug)

    if 'type' in request.args:
        schema_name = request.args.get('type')
        if '/' in schema_name:
            _, schema_name = schema_name.rsplit('/', 1)
        q = q.join(Schema)
        q = q.filter(Schema.name == schema_name)

    q = q.distinct()
    q = q.limit(get_limit(default=5))

    matches = []
    for e in q:
        data = {
            'name': e['name'].value,
            'n:type': {
                'id': '/' + project.slug + '/' + e.schema.name,
                'name': e.schema.label
            },
            'uri': url_for('entities_api.view', id=e.id, _external=True),
            'id': e.id
        }

        data['type'] = [data.get('n:type')]
        matches.append(data)

    return jsonify({
        "code": "/api/status/ok",
        "status": "200 OK",
        "prefix": request.args.get('prefix', ''),
        "result": matches
    })


@blueprint.route('/api/1/projects/<slug>/suggest_property', methods=['GET', 'POST'])
def suggest_property(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))

    prefix = '%%%s%%' % request.args.get('prefix', '')
    log.info("Suggesting property names in %s: %r", project.slug, prefix)

    q = db.session.query(Attribute)
    q = q.join(Schema)
    q = q.filter(Schema.obj == 'entity')
    q = q.filter(Schema.project == project)
    q = q.filter(or_(Attribute.label.ilike(prefix),
                     Attribute.name.ilike(prefix)))
    q = q.limit(get_limit(default=5))

    matches = []
    for attribute in q:
        matches.append({
            'name': attribute.label,
            'n:type': {
                'id': '/properties/property',
                'name': 'Property'
            },
            'id': attribute.name
        })
    return jsonify({
        "code": "/api/status/ok",
        "status": "200 OK",
        "prefix": request.args.get('prefix', ''),
        "result": matches
    })


@blueprint.route('/api/1/projects/<slug>/suggest_type', methods=['GET', 'POST'])
def suggest_type(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))

    prefix = '%%%s%%' % request.args.get('prefix', '')
    log.info("Suggesting types in %s: %r", project.slug, prefix)

    q = db.session.query(Schema)
    q = q.filter(Schema.obj == 'entity')
    q = q.filter(Schema.hidden == False) # noqa
    q = q.filter(Schema.project == project)
    q = q.filter(or_(Schema.label.ilike(prefix), Schema.name.ilike(prefix)))
    q = q.limit(get_limit(default=5))

    matches = []
    for schema in q:
        matches.append({
            'name': schema.label,
            'id': '/%s/%s' % (slug, schema.name)
        })
    return jsonify({
        "code": "/api/status/ok",
        "status": "200 OK",
        "prefix": request.args.get('prefix', ''),
        "result": matches
    })



from flask import Blueprint, request
from restpager import Pager
from sqlalchemy import or_, not_

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, request_data
from grano.model import Account
from grano.logic import accounts
from grano.core import db
from grano.views.cache import validate_cache
from grano import authz


blueprint = Blueprint('accounts_api', __name__)


@blueprint.route('/api/1/accounts/_suggest', methods=['GET'])
def suggest():
    authz.require(authz.logged_in())
    query = request.args.get('q', '') + '%'
    q = db.session.query(Account)
    q = q.filter(or_(Account.full_name.ilike(query),
                     Account.login.ilike(query),
                     Account.email.ilike(query)))
    excluded = request.args.getlist('exclude')
    if len(excluded):
        q = q.filter(not_(Account.id.in_(excluded)))
    pager = Pager(q)

    def convert(accounts):
        data = []
        for account in accounts:
            data.append({
                'display_name': account.display_name,
                'id': account.id
            })
        return data

    validate_cache(keys='#'.join([d.display_name for d in pager]))
    return jsonify(pager.to_dict(results_converter=convert))


@blueprint.route('/api/1/accounts/<id>', methods=['GET'])
def view(id):
    account = object_or_404(Account.by_id(id))
    return jsonify(account)


@blueprint.route('/api/1/accounts/<id>', methods=['POST', 'PUT'])
def update(id):
    account = object_or_404(Account.by_id(id))
    authz.require(account.id == request.account.id)
    data = request_data()
    account = accounts.save(data, account=account)
    db.session.commit()
    return jsonify(account)



from grano.core import db, url_for
from grano.model.common import UUIDBase
from grano.model.property import Property, PropertyBase


class Entity(db.Model, UUIDBase, PropertyBase):
    __tablename__ = 'grano_entity'

    same_as = db.Column(db.Unicode, db.ForeignKey('grano_entity.id'),
                        nullable=True)
    project_id = db.Column(db.Integer, db.ForeignKey('grano_project.id'))
    author_id = db.Column(db.Integer, db.ForeignKey('grano_account.id'))
    schema_id = db.Column(db.Integer, db.ForeignKey('grano_schema.id'),
                          index=True)

    degree_in = db.Column(db.Integer)
    degree_out = db.Column(db.Integer)
    degree = db.Column(db.Integer)

    inbound = db.relationship('Relation', lazy='dynamic', backref='target',
                              primaryjoin='Entity.id==Relation.target_id',
                              cascade='all, delete, delete-orphan')
    outbound = db.relationship('Relation', lazy='dynamic', backref='source',
                               primaryjoin='Entity.id==Relation.source_id',
                               cascade='all, delete, delete-orphan')

    properties = db.relationship(Property, backref='entity',
                                 order_by=Property.created_at.desc(),
                                 cascade='all, delete, delete-orphan',
                                 lazy='joined')

    @property
    def names(self):
        return [p for p in self.properties if p.name == 'name']

    @classmethod
    def by_name(cls, project, name, only_active=False):
        q = cls.by_name_many(project, name, only_active=only_active)
        return q.first()

    @classmethod
    def by_name_many(cls, project, name, only_active=False):
        q = db.session.query(cls)
        q = q.filter(cls.same_as == None) # noqa
        q = q.filter(cls.project == project)
        q = cls._filter_property(q, 'name', name, only_active=only_active)
        return q

    def to_dict_index(self):
        """ Convert an entity to the REST API form. """
        data = {
            'id': self.id,
            'degree': self.degree,
            'degree_in': self.degree_in,
            'degree_out': self.degree_out,
            'project': self.project.to_dict_short(),
            'schema': self.schema.to_dict_index(),
            'api_url': url_for('entities_api.view', id=self.id),
            'properties': {}
        }

        for prop in self.active_properties:
            name, prop = prop.to_dict_kv()
            data['properties'][name] = prop

        if self.same_as:
            data['same_as'] = self.same_as
            data['same_as_url'] = url_for('entities_api.view', id=self.same_as)
        return data

    def to_dict(self):
        """ Full serialization of the entity. """
        data = self.to_dict_index()
        data['created_at'] = self.created_at
        data['updated_at'] = self.updated_at

        if data['degree_in'] > 0:
            data['inbound_url'] = url_for('relations_api.index', target=self.id)

        if data['degree_out'] > 0:
            data['outbound_url'] = url_for('relations_api.index', source=self.id)
        return data

    def to_index(self):
        """ Convert an entity to a form appropriate for search indexing. """
        data = self.to_dict()

        data['names'] = []
        for prop in self.properties:
            if prop.name == 'name':
                data['names'].append(prop.value)

        return data



from grano.core import db
from grano.interface import Periodic, Startup


CREATE_VIEW = """
DROP MATERIALIZED VIEW IF EXISTS grano_relation_bidi;

CREATE MATERIALIZED VIEW grano_relation_bidi
    (created_at, updated_at, id, relation_id, schema_id, source_id, target_id,
        author_id, project_id, reverse)
        AS
        SELECT created_at, updated_at, id AS id, id AS relation_id, schema_id,
            source_id, target_id, author_id, project_id,
            'false'::boolean AS reverse FROM grano_relation
        UNION
        SELECT created_at, updated_at, CONCAT(id, '-rev') as id,
            id AS relation_id, schema_id, target_id AS source_id,
            source_id AS target_id, author_id, project_id,
            'true'::boolean AS reverse FROM grano_relation
    WITH DATA;
"""

REFRESH_VIEW = """
    REFRESH MATERIALIZED VIEW grano_relation_bidi;
"""


class GenerateBidi(Periodic, Startup):

    def configure(self, manager):
        db.session.execute(CREATE_VIEW)
        db.session.commit()

    def run(self):
        db.session.execute(REFRESH_VIEW)
        db.session.commit()



from unicodecsv import DictReader


class CSVImporter(object):
    """ A CSV-backed resource with the datas in it. """

    def __init__(self, fh):
        self.reader = DictReader(fh)
        self.data = list(self.reader)

    @property
    def headers(self):
        headers = set()
        for row in self.data:
            headers = headers.union(row.keys())
        return headers

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return self.data.__iter__()




"""add file foreign key to property

Revision ID: 25bb7c68b5fa
Revises: 383da0877c75
Create Date: 2014-07-14 15:09:01.525766

"""

# revision identifiers, used by Alembic.
revision = '25bb7c68b5fa'
down_revision = '992aea84727'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('grano_property', sa.Column('value_file_id', sa.Integer(), sa.ForeignKey('grano_file.id'), nullable=True))


def downgrade():
    op.drop_column('grano_property', 'value_file_id')



from time import time
import logging

from colander import Invalid
from flask import request
from werkzeug.exceptions import HTTPException

from grano.core import app
from grano.lib.serialisation import jsonify
from grano.views.base_api import blueprint as base_api
from grano.views.entities_api import blueprint as entities_api
from grano.views.relations_api import blueprint as relations_api
from grano.views.properties_api import blueprint as properties_api
from grano.views.schemata_api import blueprint as schemata_api
from grano.views.sessions_api import blueprint as sessions_api
from grano.views.projects_api import blueprint as projects_api
from grano.views.accounts_api import blueprint as accounts_api
from grano.views.files_api import blueprint as files_api
from grano.views.imports_api import blueprint as imports_api
from grano.views.pipelines_api import blueprint as pipelines_api
from grano.views.log_entries_api import blueprint as log_entries_api
from grano.views.permissions_api import blueprint as permissions_api
from grano.views.reconcile_api import blueprint as reconcile_api
from grano.views.auth import check_auth


log = logging.getLogger(__name__)


@app.errorhandler(401)
@app.errorhandler(403)
@app.errorhandler(404)
@app.errorhandler(410)
@app.errorhandler(500)
def handle_exceptions(exc):
    if isinstance(exc, HTTPException):
        message = exc.get_description(request.environ)
        message = message.replace('<p>', '').replace('</p>', '')
        body = {
            'status': exc.code,
            'name': exc.name,
            'message': message
        }
        headers = exc.get_headers(request.environ)
    else:
        body = {
            'status': 500,
            'name': exc.__class__.__name__,
            'message': unicode(exc)
        }
        headers = {}
    return jsonify(body, status=body.get('status'),
                   headers=headers)


@app.errorhandler(Invalid)
def handle_invalid(exc):
    body = {
        'status': 400,
        'name': 'Invalid Data',
        'message': unicode(exc),
        'errors': exc.asdict()
    }
    return jsonify(body, status=400)


@app.before_request
def begin_timing():
    request._begin_time = time()


@app.after_request
def end_timing(response):
    if app.config.get('DEBUG_TIMING'):
        duration = (time() - request._begin_time) * 1000
        log.debug('Request to \'%s\' (args: %r) took: %dms',
                  request.path, request.args.items(), duration)
    return response

app.register_blueprint(base_api)
app.register_blueprint(entities_api)
app.register_blueprint(relations_api)
app.register_blueprint(properties_api)
app.register_blueprint(schemata_api)
app.register_blueprint(sessions_api)
app.register_blueprint(projects_api)
app.register_blueprint(accounts_api)
app.register_blueprint(files_api)
app.register_blueprint(permissions_api)
app.register_blueprint(imports_api)
app.register_blueprint(pipelines_api)
app.register_blueprint(log_entries_api)
app.register_blueprint(reconcile_api)



"""simplify property model

Revision ID: 5143617349b9
Revises: 992aea84727
Create Date: 2014-07-22 14:15:07.069896

"""

# revision identifiers, used by Alembic.
revision = '5143617349b9'
down_revision = '25bb7c68b5fa'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_column('grano_property', 'obj')


def downgrade():
    op.add_column('grano_property',
                  sa.Column('obj', sa.VARCHAR(length=20), nullable=True))



"""empty message

Revision ID: 48dd6f2e0784
Revises: 4fb92f3ede9e
Create Date: 2014-03-14 15:16:00.950760

"""

# revision identifiers, used by Alembic.
revision = '48dd6f2e0784'
down_revision = '4fb92f3ede9e'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    op.rename_table('account', 'grano_account')
    op.rename_table('project', 'grano_project')
    op.rename_table('permission', 'grano_permission')
    op.rename_table('schema', 'grano_schema')
    op.rename_table('entity', 'grano_entity')
    op.rename_table('relation', 'grano_relation')
    op.rename_table('attribute', 'grano_attribute')
    op.rename_table('property', 'grano_property')
    op.rename_table('entity_schema', 'grano_entity_schema')
    

def downgrade():
    op.rename_table('grano_account', 'grano_account')
    op.rename_table('grano_project', 'project')
    op.rename_table('grano_permission', 'permission')
    op.rename_table('grano_schema', 'schema')
    op.rename_table('grano_entity', 'entity')
    op.rename_table('grano_relation', 'relation')
    op.rename_table('grano_attribute', 'attribute')
    op.rename_table('grano_property', 'property')
    op.rename_table('grano_entity_schema', 'entity_schema')



from hashlib import sha1

from werkzeug.http import is_resource_modified
from flask import Response, request

from grano.core import app


class NotModified(Exception):
    pass


@app.errorhandler(NotModified)
def handle_not_modified(exc):
    return Response(status=304)


def generate_etag(keys):
    args = sorted(set(request.args.items() + keys.items()))
    # jquery where is your god now?!?
    args = filter(lambda (k, v): k != '_', args)
    args = [k + ':' + repr(v) for k, v in args]
    return sha1('|'.join(args)).hexdigest()


def validate_cache(keys=None, last_modified=None):
    if isinstance(last_modified, (list, tuple)):
        last_modified = max(last_modified)
    if keys is None or not isinstance(keys, dict):
        keys = {'keys': keys}
    keys['last_modified'] = last_modified
    request._grano_etag = generate_etag(keys)
    request._grano_modified = last_modified
    if not is_resource_modified(request.environ,
                                etag=request._grano_etag,
                                last_modified=last_modified):
        raise NotModified()
    if request.if_none_match == request._grano_etag:
        raise NotModified()
    request.no_cache = False


def disable_cache():
    request.no_cache = True


@app.before_request
def setup_cache():
    request._grano_etag = None
    request._grano_modified = None
    request.no_cache = True


@app.after_request
def configure_cache_headers(response_class):
    if (not app.config.get('CACHE')) or request.no_cache:
        return response_class
    if request.method in ['GET', 'HEAD', 'OPTIONS'] \
            and response_class.status_code < 400 \
            and not response_class.is_streamed:

        response_class.cache_control.max_age = app.config.get('CACHE_AGE')
        response_class.cache_control.must_revalidate = True
        if request.account is not None:
            response_class.cache_control.private = True
        else:
            response_class.cache_control.public = True
        if request._grano_modified:
            response_class.last_modified = request._grano_modified
        if request._grano_etag:
            response_class.add_etag(request._grano_etag)
    return response_class


@app.after_request
def add_cors_header(response):
    if request.method in ('GET', 'HEAD', 'OPTIONS'):
        response.headers['Access-Control-Allow-Origin'] = '*'
    return response



from grano.core import db, url_for
from grano.model.util import make_token
from grano.model.common import IntBase


class Account(db.Model, IntBase):
    __tablename__ = 'grano_account'

    github_id = db.Column(db.Unicode)
    twitter_id = db.Column(db.Unicode)
    facebook_id = db.Column(db.Unicode)

    full_name = db.Column(db.Unicode)
    login = db.Column(db.Unicode)
    email = db.Column(db.Unicode)
    api_key = db.Column(db.Unicode, default=make_token)
    
    projects = db.relationship('Project', backref='author', lazy='dynamic')
    pipelines = db.relationship('Pipeline', backref='author', lazy='dynamic')
    files = db.relationship('File', backref='author', lazy='dynamic')
    properties = db.relationship('Property', backref='author', lazy='dynamic')
    relations = db.relationship('Relation', backref='author', lazy='dynamic')
    entities = db.relationship('Entity', backref='author', lazy='dynamic')
    permissions = db.relationship('Permission', backref='account', lazy='dynamic',
                                  cascade='all, delete, delete-orphan')
    
    @property
    def display_name(self):
        if self.full_name is not None and len(self.full_name.strip()):
            return self.full_name
        if self.login is not None and len(self.login.strip()):
            return self.login
        return self.email

    @classmethod
    def by_api_key(cls, api_key):
        q = db.session.query(cls).filter_by(api_key=api_key)
        return q.first()


    @classmethod
    def by_login(cls, login):
        q = db.session.query(cls).filter_by(login=login)
        return q.first()


    @classmethod
    def by_github_id(cls, github_id):
        q = db.session.query(cls).filter_by(github_id=str(github_id))
        return q.first()

    @classmethod
    def by_twitter_id(cls, twitter_id):
        q = db.session.query(cls).filter_by(twitter_id=str(twitter_id))
        return q.first()

    @classmethod
    def by_facebook_id(cls, facebook_id):
        q = db.session.query(cls).filter_by(facebook_id=str(facebook_id))
        return q.first()

    def to_dict_index(self):
        return {
            'id': self.id,
            'api_url': url_for('accounts_api.view', id=self.id),
            'display_name': self.display_name
            }

    def to_dict(self):
        data = self.to_dict_index()
        data['login'] = self.login
        data['full_name'] = self.full_name
        data['github_id'] = self.github_id
        data['twitter_id'] = self.twitter_id
        data['facebook_id'] = self.facebook_id
        data['created_at'] = self.created_at 
        data['updated_at'] = self.updated_at
        data['email'] = self.email
        return data




"""transform the account

Revision ID: 3c486a5dcceb
Revises: 2b65d3dfa1de
Create Date: 2014-02-25 14:09:45.192072

"""

# revision identifiers, used by Alembic.
revision = '3c486a5dcceb'
down_revision = '2b65d3dfa1de'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.add_column('account', sa.Column('facebook_id', sa.Unicode(), nullable=True))
    op.add_column('account', sa.Column('full_name', sa.Unicode(), nullable=True))
    op.add_column('account', sa.Column('twitter_id', sa.Unicode(), nullable=True))
    op.create_index('ix_property_entity_id', 'property', ['entity_id'], unique=False)
    op.create_index('ix_property_name', 'property', ['name'], unique=False)
    op.create_index('ix_property_relation_id', 'property', ['relation_id'], unique=False)
    op.create_index('ix_relation_schema_id', 'relation', ['schema_id'], unique=False)
    op.create_index('ix_relation_source_id', 'relation', ['source_id'], unique=False)
    op.create_index('ix_relation_target_id', 'relation', ['target_id'], unique=False)
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_index('ix_relation_target_id', table_name='relation')
    op.drop_index('ix_relation_source_id', table_name='relation')
    op.drop_index('ix_relation_schema_id', table_name='relation')
    op.drop_index('ix_property_relation_id', table_name='property')
    op.drop_index('ix_property_name', table_name='property')
    op.drop_index('ix_property_entity_id', table_name='property')
    op.drop_column('account', 'twitter_id')
    op.drop_column('account', 'full_name')
    op.drop_column('account', 'facebook_id')
    ### end Alembic commands ###



from grano.core import db
from grano.model import Account


def save(data, account=None):
    if account is None:
        account = Account()
        account.github_id = data.get('github_id')
        account.twitter_id = data.get('twitter_id')
        account.facebook_id = data.get('facebook_id')
    account.login = data.get('login')

    if data.get('full_name'):
        account.full_name = data.get('full_name')

    if data.get('email'):
        account.email = data.get('email')

    db.session.add(account)
    db.session.flush()
    return account


def console_account(login='_system'):
    """ Create the default system account. """
    account = Account.by_login(login)
    if account is None:
        account = save({
            'login': login,
            'email': None,
            'full_name': 'System Operations'
            })
    return account



from flask import Blueprint, request
from restpager import Pager

from grano.lib.serialisation import jsonify
from grano.model import Property, Relation, Entity
from grano.views.cache import validate_cache
from grano.lib.args import arg_bool
from grano import authz


blueprint = Blueprint('properties_api', __name__)


@blueprint.route('/api/1/relations/<obj_id>/properties', methods=['GET'])
def relations_index(obj_id):
    query = Property.all()
    query = query.filter(Property.relation_id != None)
    obj = Relation.by_id(obj_id)
    query = query.filter_by(relation_id=obj_id)
    return _index(query, obj)


@blueprint.route('/api/1/entities/<obj_id>/properties', methods=['GET'])
def entities_index(obj_id):
    query = Property.all()
    query = query.filter(Property.relation_id != None)
    obj = Entity.by_id(obj_id)
    query = query.filter_by(entity_id=obj_id)
    return _index(query, obj)


def _index(query, obj):
    authz.require(authz.project_read(obj.project))
    active_only = arg_bool('active', default=True)
    if active_only:
        query = query.filter_by(active=True)

    if 'name' in request.args:
        query = query.filter_by(name=request.args.get('name'))

    query = query.order_by(Property.created_at.desc())
    pager = Pager(query, obj_id=obj.id)
    validate_cache(keys=pager.cache_keys())
    return jsonify(pager, index=False)



# coding: utf-8
from StringIO import StringIO
import unicodecsv

from grano.logic import Loader
from grano.logic.schemata import import_schema


# This source URL will be applied to all properties without their own lineage:
DEFAULT_SOURCE_URL = 'http://www.opennews.org/'

DATA = """fellow_name,twitter_handle,start_date,end_date,organization_name,organization_url
Ben Chartoff,,2014-02-01,2014-12-01,Washington Post,http://www.washingtonpost.com/
Mark Boas,maboas,2012-02-01,2012-12-01,Al Jazeera English,http://www.aljazeera.com/
Noah Veltman,veltman,2013-02-01,2013-12-01,BBC,http://www.bbc.co.uk
Laurian Gridinoc,gridinoc,2012-02-01,2012-12-01,BBC,http://www.bbc.co.uk
Sonya Song,sonya2song,2013-02-01,2013-12-01,Boston Globe,http://www.bostonglobe.com/
Dan Schultz,slifty,2012-02-01,2012-12-01,Boston Globe,http://www.bostonglobe.com/
Gabriela Rodriguez,gaba,2014-02-01,2014-12-01,La Nacion,http://www.lanacion.com
Manuel Aristarán,manuelaristaran,2013-02-01,2013-12-01,La Nacion,http://www.lanacion.com
Harlo Holmes,harlo,2014-02-01,2014-12-01,New York Times,http://www.nytimes.com
Brian Abelson,brianabelson,2013-02-01,2013-12-01,New York Times,http://www.nytimes.com
Brian Jacobs,btjakes,2014-02-01,2014-12-01,ProPublica,http://www.propublica.org
Mike Tigas,mtigas,2013-02-01,2013-12-01,ProPublica,http://www.propublica.org
Friedrich Lindenberg,pudo,2013-02-01,2013-12-01,Spiegel Online,http://www.spiegel.de/
Marcos Vanetta,malev,2014-02-01,2014-12-01,Texas Tribune,http://www.texastribune.org/
Stijn Debrouwere,stdbrouw,2013-02-01,2013-12-01,The Guardian,http://www.theguardian.com/uk
Nicola Hughes,DataMinerUK,2012-02-01,2012-12-01,The Guardian,http://www.theguardian.com/uk
Aurelia Moser,auremoser,2014-02-01,2014-12-01,Ushahidi / Internews Kenya,http://www.ushahidi.com/
Annabel Church,annabelchurch,2013-02-01,2013-12-01,Zeit Online,http://www.zeit.de/index
Cole Gillespie,theCole,2012-02-01,2012-12-01,Zeit Online,http://www.zeit.de/index"""

SCHEMATA = """
- name: fellow
  label: An OpenNews fellow
  obj: entity
  hidden: no
  attributes:
    - name: twitter_handle
      label: Twitter handle

- name: news_organization
  label: A news organization
  obj: entity
  hidden: no
  attributes:
    - name: url
      label: URL

- name: fellowship
  label: A Fellowship
  label_in: Was hosted by
  label_out: Worked for
  obj: relation
  attributes:
    - name: start_date
      label: Start date
    - name: end_date
      label: End date
"""


def create_fixtures():
    loader = Loader('opennews', project_label='Open News',
                    project_settings={},
                    source_url=DEFAULT_SOURCE_URL)
    
    import_schema(loader.project, StringIO(SCHEMATA))

    reader = unicodecsv.reader(StringIO(DATA))
    reader.next()
     
    for record in reader:
        fellow = loader.make_entity(['fellow'])
        fellow.set('name', record[0])
        fellow.set('twitter_handle', record[1])
        fellow.save()
     
        news_org = loader.make_entity(['news_organization'])
        news_org.set('name', record[4])
        news_org.set('url', record[5])
        news_org.save()
     
        fellowship = loader.make_relation('fellowship', fellow, news_org)
        fellowship.set('start_date', record[2])
        fellowship.set('end_date', record[3])
        fellowship.save()

    loader.persist()


def create_loader(project_slug, source_url=None, project_label=None):
    return Loader(project_slug,
                  project_label=project_label or project_slug,
                  project_settings={},
                  source_url=source_url or DEFAULT_SOURCE_URL)



import logging
from stevedore.enabled import EnabledExtensionManager

from grano.core import app


log = logging.getLogger(__name__)

NAMESPACES = [
    'grano.startup',
    'grano.periodic',
    'grano.entity.change',
    'grano.relation.change',
    'grano.schema.change',
    'grano.project.change'
]
PLUGINS = {'LOADED': False, 'MANAGERS': {}}


def _get_manager(namespace):
    default_plugins = app.config.get('DEFAULT_PLUGINS', [])
    enabled_plugins = app.config.get('PLUGINS', []) + default_plugins
    available_plugins = set()

    assert namespace in NAMESPACES, \
        '%s not one of %r' % (namespace, NAMESPACES)

    if not PLUGINS['LOADED']:

        for namespace_ in NAMESPACES:
            def check_func(ext):
                available_plugins.add(ext.name)
                return ext.name in enabled_plugins

            mgr = EnabledExtensionManager(
                namespace=namespace_,
                check_func=check_func,
                propagate_map_exceptions=False,
                invoke_on_load=True)
            PLUGINS['MANAGERS'][namespace_] = mgr

        PLUGINS['LOADED'] = True

        log.info("Enabled: %s", ", ".join(sorted(enabled_plugins)))
        log.info("Available: %s", ", ".join(sorted(available_plugins)))

    return PLUGINS['MANAGERS'][namespace]


def notify_plugins(namespace, callback):
    """ Notify all active plugins about an event in a given namespace.
    The second argument is a function, it'll be called once with each
    plugin instance which is available. """

    try:
        mgr = _get_manager(namespace)
        mgr.map(lambda ext, data: callback(ext.obj), None)
    except RuntimeError:
        pass


def list_plugins():
    """ List all available plugins, grouped by the namespace in which
    they're made available. """

    plugins = {}

    for namespace in NAMESPACES:
        plugins[namespace] = []
        mgr = _get_manager(namespace)
        try:
            mgr.map(lambda e, d: plugins[namespace].append(e.name), None)
        except RuntimeError:
            pass

    return plugins



import colander
from colander import Invalid


class All(object):
    """ Composite validator which succeeds if none of its
    subvalidators raises an :class:`colander.Invalid` exception"""
    def __init__(self, *validators):
        self.validators = validators

    def __call__(self, node, value):
        for validator in self.validators:
            validator(node, value)


FORBIDDEN = ['project', 'source', 'target', 'id', 'created_at', 'updated_at',
             'author', 'author_id', 'entity', 'relation', 'entities'
             'relations', 'project', 'projects', 'grano', 'account',
             'accounts', 'graph', 'attribute', 'attributes', 'schema',
             'schemata', 'import', 'imports', 'property', 'properties']
database_format = colander.Regex(r'^[a-zA-Z][a-zA-Z0-9_-]+[a-zA-Z0-9]$')
database_forbidden = colander.Function(lambda t: t not in FORBIDDEN)
database_name = All(database_format, database_forbidden)


class FixedValue(object):
    def __init__(self, value):
        self.value = value

    def serialize(self, node, appstruct):
        return colander.null

    def deserialize(self, node, cstruct):
        return self.value

    def cstruct_children(self, node, cstruct):
        return []



"""pipeline model improved

Revision ID: 2372546f7078
Revises: 4d7168864daa
Create Date: 2014-04-07 22:05:01.431021

"""

# revision identifiers, used by Alembic.
revision = '2372546f7078'
down_revision = '4d7168864daa'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

def upgrade():
    op.add_column('grano_pipeline', sa.Column('label', sa.Unicode(), nullable=True))
    op.drop_column('grano_log_entry', 'level')
    op.add_column('grano_log_entry', sa.Column('level', sa.Integer(), nullable=True))
    

def downgrade():
    op.drop_column('grano_pipeline', 'title')
    



# This file exists because I'm not sure we'll need to subclass them later.
from werkzeug.exceptions import HTTPException, NotFound, Gone
from werkzeug.exceptions import BadRequest, Unauthorized, Forbidden
from werkzeug.exceptions import InternalServerError as InternalError
from werkzeug.exceptions import NotImplemented



"""introduce private project and roles

Revision ID: 2b269c4557de
Revises: 3c486a5dcceb
Create Date: 2014-02-25 23:16:30.479004

"""

# revision identifiers, used by Alembic.
revision = '2b269c4557de'
down_revision = '3c486a5dcceb'

from alembic import op
import sqlalchemy as sa


def upgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.create_table('role',
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('reader', sa.Boolean(), nullable=True),
    sa.Column('editor', sa.Boolean(), nullable=True),
    sa.Column('admin', sa.Boolean(), nullable=True),
    sa.Column('account_id', sa.Integer(), nullable=True),
    sa.Column('project_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['account_id'], ['account.id'], ),
    sa.ForeignKeyConstraint(['project_id'], ['project.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    op.add_column(u'project', sa.Column('private', sa.Boolean(), nullable=True))
    ### end Alembic commands ###


def downgrade():
    ### commands auto generated by Alembic - please adjust! ###
    op.drop_column(u'project', 'private')
    op.drop_table('role')
    ### end Alembic commands ###



import os
import pkg_resources

from flask import Flask, url_for as flask_url_for
from flask_oauth import OAuth
from flask_sqlalchemy import SQLAlchemy
from flask_migrate import Migrate
from kombu import Exchange, Queue
from celery import Celery

from grano import default_settings
from grano import constants, logs # noqa


app = Flask(__name__)
app.config.from_object(default_settings)
app.config.from_object(constants)
app.config.from_envvar('GRANO_SETTINGS', silent=True)

app_name = app.config.get('APP_NAME', 'grano')
app_version = pkg_resources.require('grano')[0].version

db = SQLAlchemy(app)

ALEMBIC_DIR = os.path.join(os.path.dirname(__file__), 'alembic')
ALEMBIC_DIR = os.path.abspath(ALEMBIC_DIR)
migrate = Migrate(app, db, directory=ALEMBIC_DIR)

queue_name = app.config.get('CELERY_APP_NAME') + '_q'
app.config['CELERY_DEFAULT_QUEUE'] = queue_name
app.config['CELERY_QUEUES'] = (
    Queue(queue_name, Exchange(queue_name), routing_key=queue_name),
)


celery = Celery(app.config.get('CELERY_APP_NAME', app_name),
                broker=app.config['CELERY_BROKER_URL'])
celery.config_from_object(app.config)

oauth = OAuth()


def url_for(*a, **kw):
    """Generate external URLs with HTTPS (if configured)."""
    try:
        kw['_external'] = True
        if app.config.get('PREFERRED_URL_SCHEME'):
            kw['_scheme'] = app.config.get('PREFERRED_URL_SCHEME')
        return flask_url_for(*a, **kw)
    except RuntimeError:
        return None



import logging

from grano.model import Entity, Relation, Project
from grano.logic.entities import _entity_changed
from grano.logic.relations import _relation_changed
from grano.logic.projects import _project_changed
from grano.logic.schemata import _schema_changed


log = logging.getLogger(__name__)


def rebuild():
    """ Execute the change processing handlers for all entities and
    relations currently loaded. This can be used as a housekeeping
    function. """

    for project in Project.all():
        _project_changed(project.slug, 'delete')
        _project_changed(project.slug, 'create')

        for schema in project.schemata:
            _schema_changed(schema.project.slug, schema.name, 'delete')
            _schema_changed(schema.project.slug, schema.name, 'create')

        eq = Entity.all().filter_by(same_as=None)
        eq = eq.filter_by(project=project)
        for i, entity in enumerate(eq):
            if i > 0 and i % 1000 == 0:
                log.info("Rebuilt: %s entities", i)
            _entity_changed(entity.id, 'delete')
            _entity_changed(entity.id, 'create')

        rq = Relation.all().filter_by(project=project)
        for i, relation in enumerate(rq):
            if i > 0 and i % 1000 == 0:
                log.info("Rebuilt: %s relation", i)
            _relation_changed(relation.id, 'delete')
            _relation_changed(relation.id, 'create')



from StringIO import StringIO
import json

from flask import Blueprint, request
from flask import send_file
from restpager import Pager

from grano.lib.serialisation import jsonify
from grano.lib.args import object_or_404, request_data
from grano.model import Project
from grano.logic import projects
from grano.logic.aliases import export_aliases
from grano.lib.exc import Gone, BadRequest
from grano.core import db
from grano.query import run_query
from grano.views.cache import validate_cache
from grano import authz


blueprint = Blueprint('projects_api', __name__)


@blueprint.route('/api/1/projects', methods=['GET'])
def index():
    q = Project.all()
    q = q.filter(Project.id.in_(authz.permissions().get('reader')))
    pager = Pager(q)
    validate_cache(keys=pager.cache_keys())
    return jsonify(pager, index=True)


@blueprint.route('/api/1/projects', methods=['POST', 'PUT'])
def create():
    authz.require(authz.project_create())
    project = projects.save(request_data({'author': request.account}))
    db.session.commit()
    return jsonify(project, status=201)


@blueprint.route('/api/1/projects/<slug>', methods=['GET'])
def view(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))
    if not project.private:
        validate_cache(last_modified=project.updated_at)
    return jsonify(project)


@blueprint.route('/api/1/projects/<slug>/query', methods=['GET', 'POST'])
def query(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))

    if request.method == 'POST':
        query = request.json
    else:
        try:
            query = json.loads(request.args.get('query', 'null'))
            assert query is not None
        except (TypeError, ValueError, AssertionError):
            raise BadRequest('Invalid data submitted')

    eq = run_query(project, query)
    res = {
        'status': 'ok',
        'query': eq.node,
        'results': eq.run(),
        'total': eq.count()
    }
    return jsonify(res)


@blueprint.route('/api/1/projects/<slug>/aliases', methods=['GET'])
def aliases(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_read(project))
    sio = StringIO()
    export_aliases(project, sio)
    sio.seek(0)
    res = send_file(sio, mimetype='text/csv')
    res.headers['Content-Disposition'] = 'filename=%s-aliases.csv' % project.slug
    return res


@blueprint.route('/api/1/projects/<slug>', methods=['POST', 'PUT'])
def update(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_manage(project))
    data = request_data({'author': request.account})
    project = projects.save(data, project=project)
    db.session.commit()
    return jsonify(project)


@blueprint.route('/api/1/projects/<slug>', methods=['DELETE'])
def delete(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_delete(project))
    projects.delete(project)
    db.session.commit()
    raise Gone()


@blueprint.route('/api/1/projects/<slug>/_truncate', methods=['DELETE'])
def truncate(slug):
    project = object_or_404(Project.by_slug(slug))
    authz.require(authz.project_delete(project))
    projects.truncate(project)
    db.session.commit()
    return jsonify(project)
